# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from abc import ABC, abstractmethod
from typing import Literal, Optional, Union, cast

import pandas as pd
from typing_extensions import override

from oumi.core.datasets.base_map_dataset import BaseMapDataset
from oumi.core.tokenizers import BaseTokenizer
from oumi.core.tokenizers.utils import (
    tokenize_for_completions_only_training_with_prefix,
    tokenize_for_completions_only_training_with_template,
)
from oumi.core.types.conversation import Conversation
from oumi.utils.logging import logger


class BaseSftDataset(BaseMapDataset, ABC):
    """In-memory dataset for SFT data."""

    default_dataset = None

    def __init__(
        self,
        *,
        dataset_name: Optional[str] = None,
        dataset_path: Optional[str] = None,
        split: Optional[str] = None,
        tokenizer: Optional[BaseTokenizer] = None,
        task: Literal["sft", "generation", "auto"] = "auto",
        return_tensors: bool = False,
        text_col: str = "text",
        assistant_only: bool = False,
        response_template: Optional[str] = None,
        instruction_template: Optional[str] = None,
        return_conversations: bool = False,
        return_conversations_format: Literal["dict", "json"] = "json",
        **kwargs,
    ) -> None:
        """Initializes a new instance of the BaseSftDataset class."""
        super().__init__(
            dataset_name=dataset_name,
            dataset_path=dataset_path,
            split=split,
            **kwargs,
        )

        self._task = task
        self._text_col = text_col
        self._tokenizer = tokenizer
        self._return_tensors = "pt" if return_tensors else None

        self._assistant_only = assistant_only
        self._response_template = response_template
        self._instruction_template = instruction_template
        self._return_conversations = return_conversations
        self._return_conversations_format = return_conversations_format

        if self._assistant_only:
            self._verify_assistant_only_compatibility()

        self._data = self._load_data()

    #
    # Properties
    #
    @property
    def text_col(self) -> str:
        """Gets the text target column.

        The generated text will be stored in this column.
        """
        return self._text_col

    @property
    def task(self) -> str:
        """Gets the task mode for the dataset.

        The generated prompt is often different for generation vs SFT tasks.
        """
        return self._task

    @property
    def assistant_only(self) -> bool:
        """Gets whether the dataset is set to train only on assistant turns."""
        return self._assistant_only

    #
    # Main API
    #

    def prompt(self, idx: int) -> str:
        """Returns the prompt at the specified index.

        Args:
            idx (int): The index of the prompt to retrieve.

        Returns:
            str: The prompt at the specified index.
        """
        return self.tokenize(self.conversation(idx), tokenize=False)[self.text_col]

    def conversation(self, idx: int) -> Conversation:
        """Returns the conversation at the specified index.

        Args:
            idx (int): The index of the conversation to retrieve.

        Returns:
            str: The conversation at the specified index.
        """
        sample = self.raw(idx)
        return self.transform_conversation(sample)

    def conversations(self) -> list[Conversation]:
        """Returns a list of all conversations."""
        indexes = range(len(self))
        return [self.conversation(index) for index in indexes]

    #
    # Abstract Methods
    #
    @abstractmethod
    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Preprocesses the inputs of the example and returns a dictionary.

        Args:
            example (dict): The example containing the input and instruction.

        Returns:
            dict: The preprocessed inputs as a dictionary.

        """
        raise NotImplementedError

    #
    # Pre-processing
    #
    @override
    def transform(self, sample: pd.Series) -> dict:
        """Preprocesses the inputs in the given sample."""
        conversation = self.transform_conversation(sample)
        if self._return_conversations:
            # This may require `use_torchdata=True` for TRL_SFT trainer,
            # but compatible with TRL_GRPO trainer.
            if self._return_conversations_format == "json":
                conversation_json = conversation.to_json()
                return {"conversation_json": conversation_json}
            elif self._return_conversations_format == "dict":
                return conversation.to_dict()
            else:
                raise ValueError(
                    f"Invalid return_conversations_format: "
                    f"'{self._return_conversations_format}'."
                    "Supported formats are 'json' and 'dict'."
                )
        return self.tokenize(conversation)

    def tokenize(
        self,
        sample: Union[dict, pd.Series, Conversation],
        tokenize: bool = True,
    ) -> dict:
        """Applies the chat template carried by the tokenizer to the input example.

        Args:
            sample (Dict): Mapping `messages` to a List containing the (ordered)
                messages exchanged within a single chat dialogue.
                Each item of example["messages"] is a dict mapping the `content` of the
                message and the `role` of the one relayed it.
                E.g., role == 'user' or role == 'assistant'.
            tokenize (bool): Whether to tokenize the messages or not.

        Raises:
            NotImplementedError: Currently only the `sft` task mode is supported.
            ValueError: if requested `task` is not in "sft" or "generation"

        Returns:
            Dict: It adds a `text` key in the input `example` dictionary, mapped to
            a string carrying the `messages` to the tokenizer's chat format.
        """
        if self._tokenizer is None:
            raise ValueError("Tokenizer is required for tokenization.")

        if isinstance(sample, Conversation):
            conversation = sample
        else:
            if isinstance(sample, pd.Series):
                sample = sample.to_dict()

            if isinstance(sample, dict) and "messages" in sample:
                conversation = Conversation.from_dict(sample)

            else:
                raise ValueError(
                    "Input samples must be a Conversation or a dict with "
                    "'messages' key."
                )

        if not self._assistant_only or not tokenize:
            return self._tokenize(conversation, tokenize)

        if self._is_template_compatible_with_completions_only_training:
            return tokenize_for_completions_only_training_with_template(
                tokenizer=self._tokenizer,
                conversation=conversation,
            )
        else:
            return tokenize_for_completions_only_training_with_prefix(
                tokenizer=self._tokenizer,
                conversation=conversation,
                response_template=cast(str, self._response_template),
                instruction_template=cast(str, self._instruction_template),
                response_token_ids=self.response_token_ids,
                instruction_token_ids=self.instruction_token_ids,
            )

    def _tokenize(
        self, sample: Union[dict, pd.Series, Conversation], tokenize: bool = True
    ) -> dict:
        if self._tokenizer is None:
            raise ValueError("Tokenizer is required for tokenization.")

        results = self._tokenizer.apply_chat_template(
            sample,  # type: ignore
            tokenize=tokenize,
            return_dict=tokenize,
            return_tensors=self._return_tensors,
            max_length=self._tokenizer.model_max_length,
            truncation=True,
            add_generation_prompt=(self.task == "generation"),
        )

        if tokenize:
            return cast(dict, results)
        else:
            return {
                self.text_col: results,
            }

    def _verify_assistant_only_compatibility(self) -> None:
        if self._tokenizer is None:
            raise ValueError(
                "Tokenizer is required to enable tokenization "
                "for training on assistant-only turns."
            )

        if self._tokenizer.chat_template is None:
            raise ValueError(
                "Tokenizer must have a chat template to enable "
                "tokenization for training on assistant-only turns."
            )

        template: str = self._tokenizer.chat_template  # type: ignore

        if re.search(r"\{\%-?\s*generation\s*-?\%\}", template):
            logger.info(
                "Tokenizer template contains `{% generation %}` keyword. "
                "We will use it for completions-only training."
            )

            self._is_template_compatible_with_completions_only_training = True
        else:
            if (
                self._response_template is None
                or len(self._response_template.strip()) == 0
            ):
                raise ValueError(
                    "Response template is required for completions-only training."
                )
            if self._response_template.strip() != self._response_template:
                logger.warning(
                    f"Response template '{self._response_template}' contains "
                    "leading or trailing whitespaces. These will be ignored."
                )

                self._response_template = self._response_template.strip()

            if (
                self._instruction_template is None
                or len(self._instruction_template.strip()) == 0
            ):
                raise ValueError(
                    "Instruction template is required for completions-only training."
                )

            if self._instruction_template.strip() != self._instruction_template:
                logger.warning(
                    f"Instruction template '{self._instruction_template}' contains "
                    "leading or trailing whitespaces. These will be ignored."
                )

                self._instruction_template = self._instruction_template.strip()

            self.response_token_ids = self._tokenizer.encode(
                self._response_template, add_special_tokens=False
            )

            self.instruction_token_ids = self._tokenizer.encode(
                self._instruction_template, add_special_tokens=False
            )

            self._is_template_compatible_with_completions_only_training = False



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from importlib.metadata import version
from pathlib import Path
from pprint import pformat
from typing import Any, Optional

from oumi.builders import (
    build_tuner,
)
from oumi.core.configs import (
    EvaluationConfig,
    PeftParams,
    TrainerType,
    TrainingConfig,
    TrainingParams,
    TuningConfig,
)
from oumi.core.distributed import (
    barrier,
    cleanup_distributed,
    get_device_rank_info,
    is_distributed,
    is_local_process_zero,
    is_world_process_zero,
)
from oumi.train import _ensure_dir_exists, _log_feedback_request, train
from oumi.utils.device_utils import (
    log_nvidia_gpu_runtime_info,
)
from oumi.utils.git_utils import get_git_revision_hash, get_git_tag
from oumi.utils.io_utils import save_json
from oumi.utils.logging import configure_logger, logger
from oumi.utils.torch_utils import (
    device_cleanup,
    log_devices_info,
    log_peak_gpu_memory,
    log_versioning_info,
)
from oumi.utils.version_utils import is_dev_build


def _create_tuning_dirs(config: TuningConfig) -> None:
    """Creates misc directories referenced in config."""
    _ensure_dir_exists(config.tuning.output_dir, "tuning.output_dir")
    telemetry_dir = config.tuning.telemetry_dir
    if telemetry_dir:
        _ensure_dir_exists(telemetry_dir, "training.telemetry_dir")


def _log_tuning_info(config: TuningConfig) -> None:
    """Logs misc infos about training config/devices/etc. Writes to files."""
    telemetry_dir = config.tuning.telemetry_dir
    if telemetry_dir and is_world_process_zero():
        device_rank_info = get_device_rank_info()
        save_json(
            {
                "LOCAL_WORLD_SIZE": device_rank_info.local_world_size,
                "WORLD_SIZE": device_rank_info.world_size,
            },
            telemetry_dir / "world_size.json",
        )

    if is_local_process_zero():
        log_versioning_info()
        log_devices_info(
            (telemetry_dir / "devices_info.txt")
            if telemetry_dir and is_world_process_zero()
            else None
        )
        oumi_version = version("oumi")
        logger.info(f"Oumi version: {oumi_version}")
        if is_dev_build():
            logger.info(f"Git revision hash: {get_git_revision_hash()}")
            logger.info(f"Git tag: {get_git_tag()}")


def tune(
    config: TuningConfig,
    additional_model_kwargs: Optional[dict[str, Any]] = None,
    additional_tuner_kwargs: Optional[dict[str, Any]] = None,
    verbose: bool = False,
) -> None:
    """Tunes a model using the provided configuration."""
    _START_TIME = time.time()

    _create_tuning_dirs(config)
    _log_tuning_info(config)

    # Configure logging to file
    log_dir = Path(config.tuning.output_dir) / "logs"
    for logger_name in ("oumi", "oumi.telemetry"):
        configure_logger(
            logger_name,
            level=config.tuning.log_level,
            log_dir=log_dir,
        )

    telemetry_dir = config.tuning.telemetry_dir
    if is_local_process_zero():
        if verbose:
            logger.info(f"TuningConfig:\n{pformat(config)}")
        if telemetry_dir and is_world_process_zero():
            config_path = telemetry_dir / "tuning_config.yaml"
            config.to_yaml(str(config_path))
            logger.info(f"Training config saved to {config_path}")

    # metrics_functions = build_metrics_functions(config.tuning)
    tuner = build_tuner(config.tuning)

    def objective(
        suggested_training_params: dict[str, Any],
        suggested_peft_params: dict[str, Any],
        trial_number: int,
    ) -> dict[str, float]:
        """Objective function for hyperparameter tuning."""
        # Merge suggested training params with fixed params
        training_params = {
            **config.tuning.fixed_training_params,
            **suggested_training_params,
        }
        training_params["output_dir"] = str(
            Path(
                config.tuning.output_dir,
                f"trial_{trial_number}",
            )
        )
        trial_training_params = TrainingParams(**training_params)
        trial_training_params.trainer_type = TrainerType.TRL_SFT

        # Merged suggested and fixed PEFT params
        peft_params = {
            **config.tuning.fixed_peft_params,
            **suggested_peft_params,
        }
        trial_peft_params = PeftParams(**peft_params)

        trial_train_config = TrainingConfig(
            model=config.model,
            data=config.data,
            training=trial_training_params,
            peft=trial_peft_params,
        )

        # Trains model for this trial
        # TODO: This could be a boolean, but I think for the future a dict like this is
        # good.
        eval_results = train(
            trial_train_config, additional_tuning_kwargs={"some_variable": True}
        )

        # TODO: Add better support for tuning with custom evaluations results
        #  there should be a better way to do this other then reloading
        #  everything post-training.
        custom_results = {}
        if config.tuning.custom_eval_metrics:
            # Delayed import in case the user needs custom evaluations
            from oumi.core.configs import EvaluationTaskParams
            from oumi.core.configs.params.evaluation_params import EvaluationBackend
            from oumi.core.configs.params.generation_params import GenerationParams
            from oumi.evaluate import evaluate

            task_names = config.tuning.custom_eval_metrics
            # Build EvaluationConfig mirroring training configuration
            eval_model_params = trial_train_config.model
            # Load the just-saved fine-tuned model from output_dir
            eval_model_params.model_name = trial_train_config.training.output_dir

            eval_tasks = [
                EvaluationTaskParams(
                    evaluation_backend=EvaluationBackend.CUSTOM.value,
                    task_name=task_name,
                )
                for task_name in task_names
            ]

            generation = GenerationParams(
                batch_size=(
                    trial_train_config.training.per_device_eval_batch_size or 1
                ),
                seed=trial_train_config.training.seed,
                max_new_tokens=(
                    eval_model_params.model_max_length
                    if eval_model_params.model_max_length
                    else 1024
                ),
            )

            eval_output_dir = str(
                Path(trial_train_config.training.output_dir) / "custom_eval"
            )

            eval_config = EvaluationConfig(
                tasks=eval_tasks,
                model=eval_model_params,
                generation=generation,
                run_name=trial_train_config.training.run_name,
                enable_wandb=trial_train_config.training.enable_wandb,
                output_dir=eval_output_dir,
            )

            device_cleanup()
            task_results = evaluate(eval_config)
            # Unpack result metrics
            custom_results = {
                task_name: task_result[task_name]
                for task_name, task_result in zip(task_names, task_results)
            }

        if not eval_results:
            logger.error(
                "Eval results needs to be an dictionary of metrics, "
                "for tune to work correctly"
            )
            raise

        logger.info(
            f"Trial {trial_number} finished. Evaluation results: {eval_results}"
        )

        output_metrics_dict = {}
        metrics_list = list(config.tuning.evaluation_metrics)
        if config.tuning.custom_eval_metrics:
            metrics_list.extend(config.tuning.custom_eval_metrics)
        for metric in metrics_list:
            if metric in eval_results:
                output_metrics_dict[metric] = eval_results[metric]
            elif metric in custom_results:
                output_metrics_dict[metric] = custom_results[metric]

        return output_metrics_dict

    logger.info(f"Tuning init time: {time.time() - _START_TIME:.3f}s")
    logger.info("Starting hyperparameter tuning...")
    tuner.optimize(objective, n_trials=config.tuning.n_trials)

    if len(config.tuning.evaluation_direction) == 1:
        best_trial = tuner.get_best_trial()
        logger.info(f"Best trial: {best_trial}")
    else:
        best_trials = tuner.get_best_trials()
        logger.info(f"Best trials: {pformat(best_trials)}")

    logger.info(
        "Tuning is Complete. Saving study results at "
        f"{config.tuning.output_dir}/trials_results.csv ..."
    )
    tuner.save_study(config)

    log_nvidia_gpu_runtime_info(log_prefix="GPU Metrics After Tuning:")
    log_peak_gpu_memory()
    barrier()
    if is_distributed():
        cleanup_distributed()

    _log_feedback_request()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: conversation.proto
# Protobuf Python Version: 5.27.5
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 5, 27, 5, "", "conversation.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x12\x63onversation.proto\x12\x17oumi.core.conversations"2\n\x08\x44\x61taBlob\x12\x11\n\tmime_type\x18\x01 \x01(\t\x12\x13\n\x0b\x62inary_data\x18\x02 \x01(\x0c"\xe1\x01\n\x0b\x43ontentPart\x12\x37\n\x04type\x18\x01 \x01(\x0e\x32).oumi.core.conversations.ContentPart.Type\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\t\x12/\n\x04\x62lob\x18\x03 \x01(\x0b\x32!.oumi.core.conversations.DataBlob"W\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\x08\n\x04TEXT\x10\x01\x12\x0e\n\nIMAGE_PATH\x10\x02\x12\r\n\tIMAGE_URL\x10\x03\x12\x10\n\x0cIMAGE_BINARY\x10\x04"w\n\x07Message\x12\n\n\x02id\x18\x01 \x01(\t\x12+\n\x04role\x18\x02 \x01(\x0e\x32\x1d.oumi.core.conversations.Role\x12\x33\n\x05parts\x18\x03 \x03(\x0b\x32$.oumi.core.conversations.ContentPart"\xd3\x01\n\x0c\x43onversation\x12\x17\n\x0f\x63onversation_id\x18\x01 \x01(\t\x12\x32\n\x08messages\x18\x02 \x03(\x0b\x32 .oumi.core.conversations.Message\x12\x45\n\x08metadata\x18\x03 \x03(\x0b\x32\x33.oumi.core.conversations.Conversation.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01*K\n\x04Role\x12\x14\n\x10ROLE_UNSPECIFIED\x10\x00\x12\n\n\x06SYSTEM\x10\x01\x12\x08\n\x04USER\x10\x02\x12\r\n\tASSISTANT\x10\x03\x12\x08\n\x04TOOL\x10\x04\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "conversation_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_CONVERSATION_METADATAENTRY"]._loaded_options = None
    _globals["_CONVERSATION_METADATAENTRY"]._serialized_options = b"8\001"
    _globals["_ROLE"]._serialized_start = 662
    _globals["_ROLE"]._serialized_end = 737
    _globals["_DATABLOB"]._serialized_start = 47
    _globals["_DATABLOB"]._serialized_end = 97
    _globals["_CONTENTPART"]._serialized_start = 100
    _globals["_CONTENTPART"]._serialized_end = 325
    _globals["_CONTENTPART_TYPE"]._serialized_start = 238
    _globals["_CONTENTPART_TYPE"]._serialized_end = 325
    _globals["_MESSAGE"]._serialized_start = 327
    _globals["_MESSAGE"]._serialized_end = 446
    _globals["_CONVERSATION"]._serialized_start = 449
    _globals["_CONVERSATION"]._serialized_end = 660
    _globals["_CONVERSATION_METADATAENTRY"]._serialized_start = 613
    _globals["_CONVERSATION_METADATAENTRY"]._serialized_end = 660
# @@protoc_insertion_point(module_scope)



import io
import logging
import tempfile
from pathlib import Path
from unittest.mock import call, patch

import PIL.Image
import pytest
import typer
from typer.testing import CliRunner

import oumi
from oumi.cli.alias import AliasType
from oumi.cli.cli_utils import CONTEXT_ALLOW_EXTRA_ARGS
from oumi.cli.infer import infer
from oumi.core.configs import (
    GenerationParams,
    InferenceConfig,
    InferenceEngineType,
    ModelParams,
)
from oumi.utils.logging import logger

runner = CliRunner()


@pytest.fixture
def mock_fetch():
    with patch("oumi.cli.cli_utils.resolve_and_fetch_config") as m_fetch:
        yield m_fetch


@pytest.fixture
def mock_alias():
    with patch("oumi.cli.infer.try_get_config_name_for_alias") as try_alias:
        yield try_alias


def _create_inference_config() -> InferenceConfig:
    return InferenceConfig(
        model=ModelParams(
            model_name="MlpEncoder",
            trust_remote_code=True,
            tokenizer_name="gpt2",
        ),
        generation=GenerationParams(
            max_new_tokens=5,
        ),
    )


#
# Fixtures
#
@pytest.fixture
def app():
    fake_app = typer.Typer()
    fake_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(infer)
    yield fake_app


@pytest.fixture
def mock_infer():
    with patch.object(oumi, "infer") as m_infer:
        yield m_infer


@pytest.fixture
def mock_infer_interactive():
    with patch.object(oumi, "infer_interactive") as m_infer:
        yield m_infer


def test_infer_runs(app, mock_infer, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(app, ["-i", "--config", yaml_path])
        mock_infer_interactive.assert_has_calls(
            [call(config, input_image_bytes=None, system_prompt=None)]
        )


def test_infer_with_alias_runs(app, mock_infer, mock_infer_interactive, mock_alias):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        mock_alias.return_value = yaml_path
        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(app, ["-i", "--config", "random_alias"])
        mock_alias.assert_called_once_with("random_alias", AliasType.INFER)
        mock_infer_interactive.assert_has_calls(
            [call(config, input_image_bytes=None, system_prompt=None)]
        )


def test_infer_runs_interactive_by_default(app, mock_infer, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(app, ["--config", yaml_path])
        mock_infer_interactive.assert_has_calls(
            [call(config, input_image_bytes=None, system_prompt=None)]
        )


def test_infer_fails_no_args(app, mock_infer, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)
        result = runner.invoke(app, [])
        mock_infer_interactive.assert_not_called()
        assert result.exit_code == 2


def test_infer_with_overrides(app, mock_infer, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(
            app,
            [
                "--interactive",
                "--config",
                yaml_path,
                "--model.tokenizer_name",
                "new_name",
                "--generation.max_new_tokens",
                "5",
                "--engine",
                "VLLM",
            ],
        )
        expected_config = _create_inference_config()
        expected_config.model.tokenizer_name = "new_name"
        expected_config.generation.max_new_tokens = 5
        expected_config.engine = InferenceEngineType.VLLM
        mock_infer_interactive.assert_has_calls(
            [call(expected_config, input_image_bytes=None, system_prompt=None)]
        )


def test_infer_runs_with_image(app, mock_infer, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)

        test_image = PIL.Image.new(mode="RGB", size=(32, 16))
        temp_io_output = io.BytesIO()
        test_image.save(temp_io_output, format="PNG")
        image_bytes = temp_io_output.getvalue()

        image_path = Path(output_temp_dir) / "test_image.png"
        with image_path.open(mode="wb") as f:
            f.write(image_bytes)

        _ = runner.invoke(
            app, ["-i", "--config", yaml_path, "--image", str(image_path)]
        )
        mock_infer_interactive.assert_has_calls(
            [call(config, input_image_bytes=[image_bytes], system_prompt=None)]
        )


def test_infer_not_interactive_runs(app, mock_infer, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        config: InferenceConfig = _create_inference_config()
        config.input_path = "some/path"
        config.to_yaml(yaml_path)
        _ = runner.invoke(app, ["--config", yaml_path])
        mock_infer.assert_has_calls([call(config)])


def test_infer_not_interactive_with_overrides(app, mock_infer, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        config: InferenceConfig = _create_inference_config()
        config.input_path = "some/path"
        config.to_yaml(yaml_path)
        _ = runner.invoke(
            app,
            [
                "--config",
                yaml_path,
                "--model.tokenizer_name",
                "new_name",
                "--generation.max_new_tokens",
                "5",
            ],
        )
        expected_config = _create_inference_config()
        expected_config.model.tokenizer_name = "new_name"
        expected_config.generation.max_new_tokens = 5
        expected_config.input_path = "some/path"
        mock_infer.assert_has_calls([call(expected_config)])


def test_infer_logging_levels(app, mock_infer, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")
        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(app, ["-i", "--config", yaml_path, "--log-level", "WARNING"])
        assert logger.level == logging.WARNING
        _ = runner.invoke(app, ["-i", "--config", yaml_path, "-log", "CRITICAL"])
        assert logger.level == logging.CRITICAL


def test_infer_with_system_prompt(app, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")

        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)

        # Test with interactive mode and system prompt
        result = runner.invoke(
            app,
            [
                "-i",
                "--config",
                yaml_path,
                "--system-prompt",
                "You are a mighty assistant",
            ],
        )
        assert result.exit_code == 0
        mock_infer_interactive.assert_called_once_with(
            config, system_prompt="You are a mighty assistant", input_image_bytes=None
        )
        mock_infer_interactive.reset_mock()


def test_infer_with_system_prompt_and_image(app, mock_infer_interactive):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "infer.yaml")

        config: InferenceConfig = _create_inference_config()
        config.to_yaml(yaml_path)

        test_image = PIL.Image.new(mode="RGB", size=(32, 16))
        temp_io_output = io.BytesIO()
        test_image.save(temp_io_output, format="PNG")
        image_bytes = temp_io_output.getvalue()

        image_path = Path(output_temp_dir) / "test_image.png"
        with image_path.open(mode="wb") as f:
            f.write(image_bytes)

        result = runner.invoke(
            app,
            [
                "-i",
                "--config",
                yaml_path,
                "--system-prompt",
                "You are not an average assistant",
                "--image",
                str(image_path),
            ],
        )
        assert result.exit_code == 0
        mock_infer_interactive.assert_called_once_with(
            config,
            system_prompt="You are not an average assistant",
            input_image_bytes=[image_bytes],
        )


def test_infer_with_oumi_prefix_and_explicit_output_dir(
    app, mock_fetch, mock_infer_interactive
):
    with tempfile.TemporaryDirectory() as temp_dir:
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"

        config = _create_inference_config()
        expected_path.parent.mkdir(parents=True, exist_ok=True)
        config.to_yaml(expected_path)
        mock_fetch.return_value = expected_path

        with patch.dict("os.environ", {"OUMI_DIR": str(output_dir)}):
            result = runner.invoke(app, ["-i", "--config", config_path])

        assert result.exit_code == 0
        mock_fetch.assert_called_once_with(config_path)
        mock_infer_interactive.assert_called_once_with(
            config, input_image_bytes=None, system_prompt=None
        )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Based on MFU from PaLM paper: https://arxiv.org/pdf/2204.02311."""

import time
from typing import Optional, Union

import torch
import transformers

from oumi.core.callbacks.base_trainer_callback import BaseTrainerCallback
from oumi.core.configs import TrainingParams
from oumi.core.distributed import get_device_rank_info, is_world_process_zero
from oumi.performance.mfu import calculate_mfu
from oumi.utils.logging import logger
from oumi.utils.torch_utils import get_device_name

_LOGS_KWARG = "logs"

# MFU using only the time between on_step_start and on_step_end (except the first step)
_TRAIN_STEP_MFU = "train_step_mfu"
# MFU using the time since training started (except the first step)
_TRAIN_MFU = "train_mfu"


class MfuTrainerCallback(BaseTrainerCallback):
    """Trainer callback to calculate the MFU of the model during training.

    Should be compatible with all trainers that inherit from transformers.Trainer.
    """

    def __init__(
        self,
        dtype: torch.dtype,
        num_params: int,
        sequence_length: int,
        num_layers: Optional[int] = None,
        num_attention_heads: Optional[int] = None,
        attention_head_size: Optional[int] = None,
        add_rematerialization: bool = False,
    ):
        """Initialize the MfuTrainerCallback.

        Args:
            dtype: The data type of the model.
            num_params: The number of parameters in the model.
            start_time_seconds: The start time of the program.
            sequence_length: The sequence length of the model.
            num_layers: The number of layers in the model.
            num_attention_heads: The number of attention heads in the model.
            attention_head_size: The size of each attention head in the model.
            add_rematerialization: Whether to add rematerialization to FLOPs per token.
        """
        self._dtype = dtype
        self._num_params = num_params
        self._time_of_second_step: Optional[float] = None
        self._time_for_train_steps = 0.0
        self._tokens_seen_so_far = 0
        self._sequence_length = sequence_length
        self._num_layers = num_layers
        self._num_attention_heads = num_attention_heads
        self._attention_head_size = attention_head_size
        self._add_rematerialization = add_rematerialization
        self._first_step_finished = False
        self._steps_since_last_log = 0

        device_rank_info = get_device_rank_info()
        self._num_devices = device_rank_info.world_size
        self._is_world_rank_zero = is_world_process_zero()
        logger.info(f"MFU number of devices: {self._num_devices}")
        self._device_name = get_device_name()

        logger.info(f"MFU device name: {self._device_name}")
        if self._device_name == "CPU":
            logger.warning("MFU is not supported on CPU, the callback will do nothing.")

    def _callback_disabled(self) -> bool:
        """Check if the callback should be disabled."""
        return not self._is_world_rank_zero or self._device_name == "CPU"

    def on_step_begin(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the beginning of each train step."""
        if self._callback_disabled():
            return

        self._step_start_time = time.time()
        if not self._first_step_finished:
            # Calculate the number of tokens processed per step during the first step
            self._tokens_per_step = (
                args.gradient_accumulation_steps
                * args.per_device_train_batch_size
                * self._num_devices
                * self._sequence_length
            )
            return

        if self._time_of_second_step is None:
            self._time_of_second_step = self._step_start_time

    def on_step_end(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the end of each train step.

        Note that this will be called after all gradient accumulation substeps.
        """
        if self._callback_disabled():
            return

        # Keep track of only the training step time for "ideal" MFU
        delta_time_seconds = time.time() - self._step_start_time
        if not self._first_step_finished:
            self._first_step_finished = True
            logger.info(f"First step time: {delta_time_seconds:.2f}s")
            return

        self._time_for_train_steps += delta_time_seconds
        self._steps_since_last_log += 1

    def on_log(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called after logging the last logs."""
        if self._callback_disabled():
            return

        # Avoid logging until after the first step.
        if self._time_of_second_step is None:
            return

        delta_time_seconds_train = time.time() - self._time_of_second_step
        delta_time_seconds_step = self._time_for_train_steps

        tokens_since_last_log = self._tokens_per_step * self._steps_since_last_log
        total_tokens = self._tokens_seen_so_far + tokens_since_last_log

        # MFU using only the time spent on training steps (excluding the first step).
        train_step_mfu = calculate_mfu(
            device_name=self._device_name,
            num_devices=self._num_devices,
            dtype=self._dtype,
            num_params=self._num_params,
            num_tokens=total_tokens,
            delta_time_seconds=delta_time_seconds_step,
            num_layers=self._num_layers,
            num_attention_heads=self._num_attention_heads,
            attention_head_size=self._attention_head_size,
            sequence_length=self._sequence_length,
            add_rematerialization=self._add_rematerialization,
        )
        # MFU using the time since training started (excluding the first step).
        train_mfu = calculate_mfu(
            device_name=self._device_name,
            num_devices=self._num_devices,
            dtype=self._dtype,
            num_params=self._num_params,
            num_tokens=total_tokens,
            delta_time_seconds=delta_time_seconds_train,
            num_layers=self._num_layers,
            num_attention_heads=self._num_attention_heads,
            attention_head_size=self._attention_head_size,
            sequence_length=self._sequence_length,
            add_rematerialization=self._add_rematerialization,
        )
        if _LOGS_KWARG in kwargs:
            kwargs[_LOGS_KWARG][_TRAIN_STEP_MFU] = train_step_mfu
            kwargs[_LOGS_KWARG][_TRAIN_MFU] = train_mfu

        # Cleanup values
        self._tokens_seen_so_far = total_tokens
        self._steps_since_last_log = 0



from typing import Optional

import datasets
from sklearn.metrics import balanced_accuracy_score

from oumi.core.configs import EvaluationConfig, EvaluationTaskParams
from oumi.core.evaluation import EvaluationResult
from oumi.core.inference import BaseInferenceEngine
from oumi.core.registry import register_evaluation_function
from oumi.datasets import TextSftJsonLinesDataset

_prompt_template = """You will be given a premise and a hypothesis.
Determine if the hypothesis is <|supported|> or <|unsupported|> based on the premise.

premise: {premise}

hypothesis: {hypothesis}

You are allowed to think out loud.
Ensure that your final answer is formatted as <|supported|> or <|unsupported|>.
"""


def _load_dataset(num_examples: Optional[int]) -> TextSftJsonLinesDataset:
    """Load the facebook/anli dataset, formatted as a classification problem."""
    anli_dataset = datasets.load_dataset("facebook/anli", split="test_r3")
    evaluation_dataset = []
    for anli_example in anli_dataset:
        assert isinstance(anli_example, dict)
        prompt = _prompt_template.format(
            premise=anli_example["premise"], hypothesis=anli_example["hypothesis"]
        )
        message = {"role": "user", "content": prompt}
        metadata = {"label": _convert_label_to_binary(anli_example["label"])}
        conversation = {
            "conversation_id": anli_example["uid"],
            "messages": [message],
            "metadata": metadata,
        }
        evaluation_dataset.append(conversation)

    # Limit the number of examples, if requested.
    if num_examples is not None:
        evaluation_dataset = evaluation_dataset[:num_examples]
    return TextSftJsonLinesDataset(data=evaluation_dataset)


def _convert_label_to_binary(label: int) -> int:
    """Clamps labels to the set [0,1]."""
    if label == 0:
        return 0  # supported
    elif label in [1, 2]:
        return 1  # unsupported
    else:
        raise ValueError(f"Invalid label: {label}")


def _extract_prediction(response: str) -> int:
    """Converts a response to a label: [0, 1], or -1 if inconclusive."""
    is_unsupported = "<|unsupported|>" in response
    is_supported = "<|supported|>" in response
    if is_unsupported == is_supported:
        return -1
    return 0 if is_supported else 1


@register_evaluation_function("classifier_benchmark")
def classifier_benchmark(
    task_params: EvaluationTaskParams,
    config: EvaluationConfig,
    inference_engine: BaseInferenceEngine,
):
    """Custom evaluation function registered as `my_custom_evaluation`."""
    num_examples: Optional[int] = task_params.eval_kwargs.get("num_examples", None)
    dataset = _load_dataset(num_examples)
    # Run inference to generate the model responses.
    conversations = inference_engine.infer(dataset.conversations())

    y_true, y_pred = [], []
    for conversation in conversations:
        # Extract the assistant's (LLM's) response from the conversation.
        response = conversation.last_message()
        assert response and isinstance(response.content, str)
        # Extract the prediction from the response.
        prediction = _extract_prediction(response.content)
        if prediction != -1:
            y_pred.append(prediction)
            y_true.append(conversation.metadata["label"])

    # Compute any relevant metrics (such as Balanced Accuracy).
    bacc = balanced_accuracy_score(y_true, y_pred)

    return EvaluationResult(
        task_result={
            "results": {
                "classifier_benchmark": {"alias": "classifier_benchmark", "bacc": bacc}
            }
        }
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from typing import Any, Callable

from oumi.core.configs.tuning_config import TuningConfig, TuningParams


class BaseTuner(ABC):
    """Abstract base class for hyperparameter tuners.

    This class defines the interface that all tuner implementations must follow,
    allowing for different optimization backends (Optuna, Ray Tune, etc.) while
    maintaining a consistent API.
    """

    def __init__(self, tuning_params: TuningParams):
        """Initialize the tuner with configuration parameters.

        Args:
            tuning_params: Configuration for the tuning process.
        """
        self.tuning_params = tuning_params
        self._study = None

    @abstractmethod
    def create_study(self) -> None:
        """Create a new optimization study.

        This method should initialize the tuner's internal study object
        with the appropriate configuration (study name, direction, etc.).
        """
        pass

    @abstractmethod
    def suggest_parameters(self, trial: Any) -> dict[str, Any]:
        """Suggest hyperparameters for a trial.

        Args:
            trial: The trial object from the underlying tuner backend.

        Returns:
            Dictionary mapping parameter names to suggested values.
        """
        pass

    @abstractmethod
    def optimize(
        self,
        objective_fn: Callable[..., Any],
        n_trials: int,
    ) -> None:
        """Run the optimization process.

        Args:
            objective_fn: Function that takes suggested parameters and returns
                         a dictionary of metric values.
            n_trials: Number of trials to run.

        Returns:
            None
        """
        pass

    @abstractmethod
    def get_best_trial(self) -> dict[str, Any]:
        """Get the best trial from the study, if only one objective is being optimized.

        Returns:
            Dictionary containing best parameters and their metric values.
        """
        pass

    @abstractmethod
    def get_best_trials(self) -> list[dict[str, Any]]:
        """Get the best trials from the study, for multiple objectives.

        Returns:
            Dictionary containing best parameters and their metric values for the best
            trials.
        """
        pass

    @abstractmethod
    def save_study(self, config: TuningConfig) -> None:
        """Saves the study object to the specified output directory.

        Args:
            config (TrainingConfig): The Oumi training config.

        Returns:
            None
        """
        pass



import os
import sys

from oumi.models.experimental.cambrian.constants import LOGDIR

server_error_msg = (
    "**NETWORK ERROR DUE TO HIGH TRAFFIC. PLEASE REGENERATE OR REFRESH THIS PAGE.**"
)
moderation_msg = (
    "YOUR INPUT VIOLATES OUR CONTENT MODERATION GUIDELINES. PLEASE TRY AGAIN."
)

# handler = None

# TODO: move elsewhere?
IS_XLA_AVAILABLE = False
try:
    import torch_xla

    IS_XLA_AVAILABLE = True
except ImportError:
    pass



import tempfile

from transformers import Trainer

from oumi import train
from oumi.builders.data import build_dataset_mixture
from oumi.builders.models import (
    build_model,
    build_tokenizer,
    is_image_text_llm,
)
from oumi.core.configs import (
    DataParams,
    DatasetParams,
    DatasetSplit,
    DatasetSplitParams,
    ModelParams,
    TrainerType,
    TrainingConfig,
    TrainingParams,
)
from oumi.core.configs.internal.supported_models import (
    is_custom_model,
)


def _get_default_config(output_temp_dir):
    return TrainingConfig(
        data=DataParams(
            train=DatasetSplitParams(
                datasets=[
                    DatasetParams(
                        dataset_name="debug_pretraining",
                        dataset_kwargs={"dataset_size": 25, "seq_length": 128},
                    )
                ],
                stream=True,
                pack=True,
            ),
        ),
        model=ModelParams(
            model_name="MlpEncoder",
            tokenizer_name="gpt2",
            model_max_length=128,
            trust_remote_code=False,
            load_pretrained_weights=False,
            model_kwargs={
                "input_dim": 50257,
                "output_dim": 50257,
            },  # vocab size of GPT2 tokenizer
            tokenizer_pad_token="<|endoftext|>",
        ),
        training=TrainingParams(
            trainer_type=TrainerType.HF,
            max_steps=3,
            logging_steps=1,
            enable_wandb=False,
            enable_tensorboard=False,
            enable_mlflow=False,
            output_dir=output_temp_dir,
            include_performance_metrics=False,
            include_alternative_mfu_metrics=True,
        ),
    )


def test_train_native_pt_model_from_api():
    with tempfile.TemporaryDirectory() as output_temp_dir:
        config: TrainingConfig = _get_default_config(output_temp_dir)

        assert is_custom_model(config.model.model_name), f"ModelParams: {config.model}"
        assert not is_image_text_llm(config.model), f"ModelParams: {config.model}"

        tokenizer = build_tokenizer(config.model)

        dataset = build_dataset_mixture(
            config.data,
            tokenizer,
            DatasetSplit.TRAIN,
            seq_length=config.model.model_max_length,
        )

        model = build_model(model_params=config.model)

        training_args = config.training.to_hf()

        trainer = Trainer(
            model=model,
            processing_class=tokenizer,
            args=training_args,
            train_dataset=dataset,
        )

        trainer.train()


def test_train_native_pt_model_from_config():
    with tempfile.TemporaryDirectory() as output_temp_dir:
        config = _get_default_config(output_temp_dir)

        assert is_custom_model(config.model.model_name), f"ModelParams: {config.model}"
        assert not is_image_text_llm(config.model), f"ModelParams: {config.model}"

        train(config)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Any, Optional

from oumi.core.registry import RegistryType, register


def _extract_prediction(response: str) -> Optional[int]:
    r"""Returns the numeric answer extracted from `\boxed{...}`, or None otherwise."""
    regex_result = re.findall(r"\\boxed\{([-+]?\d+)\}", response)
    if not regex_result or len(regex_result) != 1:
        return None
    number_str = regex_result[0]
    # Except clause shouldn't trigger because the regex should only find ints.
    try:
        return int(number_str)
    except ValueError:
        return None


def compute_letter_count_reward(completion: str, target_count: int) -> float:
    """Computes the rewards for counting the letters in a string.

    Args:
        completion: The completion string from the LLM.
        target_count: The target count of letters.

    Returns:
        The reward value.
    """
    count = _extract_prediction(completion)

    # Lowest reward goes to unparseable responses
    if count is None:
        return -3.0

    delta = abs(count - target_count)

    # Reward scales from [0, -2) as delta increases
    # Ensures that "worse" answers (where the counts are off by a higher amount) are
    # penalized while never reaching -3.0 which is reserved for unparseable answers.
    return (1 / (delta + 0.5)) - 2


@register("count_letters", RegistryType.REWARD_FUNCTION)
def _count_letters(
    completions: list[list[dict[str, Any]]],
    letter_count: list[int],
    **kwargs: dict[str, Any],
) -> list[float]:
    """Custom reward function for counting letters in a string.

    For more details on custom reward functions used in trl's GRPOTrainer, see:
    https://huggingface.co/docs/trl/main/en/grpo_trainer#using-a-custom-reward-function.

    Args:
        completions: The list of completions from the LLM.
        letter_count: The list of target count of letters.
        kwargs: Unused.

    Returns:
        The reward values for each completion, calculated as the negative of the
        absolute difference between the count and the target count. The count is assumed
        to be the last group of consecutive digits in the completion string.
    """
    completions_strs = [c[0]["content"] for c in completions]
    return [
        compute_letter_count_reward(c, t)
        for c, t in zip(completions_strs, letter_count)
    ]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import math
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, NamedTuple, Optional, TypeVar, Union, cast

import numpy as np
import torch

from oumi.utils.device_utils import get_nvidia_gpu_memory_utilization
from oumi.utils.logging import logger
from oumi.utils.str_utils import compute_utf8_len


def device_cleanup() -> None:
    """Empties gpu cache, good to do before and after training for cleanup."""
    logger.debug("Running garbage collection.")
    gc.collect()

    if torch.cuda.is_available():
        logger.debug("Cleaning up GPU memory.")
        logger.debug(
            "GPU memory occupied before cleanup: "
            f"{get_nvidia_gpu_memory_utilization()} MiB"
        )

        torch.cuda.empty_cache()

        logger.debug(f"Memory after cleanup: {get_nvidia_gpu_memory_utilization()} MiB")

    elif torch.backends.mps.is_available():
        logger.debug("Cleaning up MPS memory.")
        torch.mps.empty_cache()


def limit_per_process_memory(percent: float = 0.95) -> None:
    """Limits process memory by a certain percentage.

    On Windows and WSL, there's a pool of 'shared gpu memory'.
    This pool is using the RAM (slow) on one's machine rather than actual
    VRAM (fast). Setting this value ensures your machine never uses the slow
    memory and OOMs instead. Note that this may not be needed on Linux machines
    since this is an OS-level feature.
    """
    if torch.cuda.is_available():
        torch.cuda.set_per_process_memory_fraction(percent)


def format_cudnn_version(v: Optional[int]) -> str:
    """Formats the cuDNN version number.

    Args:
        v: The cuDNN version number.

    Returns:
        A formatted string.
    """
    if v is None:
        return ""
    return ".".join(map(str, (v // 1000, v // 100 % 10, v % 100)))


def log_versioning_info() -> None:
    """Logs misc versioning information."""
    logger.info(f"Torch version: {torch.__version__}. NumPy version: {np.__version__}")
    if not torch.cuda.is_available():
        logger.info("CUDA is not available!")
        return

    # pyright seems to have an issue with torch==2.5.1
    # torch.version is always available, but pyright doesn't know that
    if hasattr(torch, "version"):
        logger.info(f"CUDA version: {torch.version.cuda} ")  # type: ignore

    # For AMD GPUs, these functions return ROCm, MlOpen versions respectively.
    logger.info(
        f"CuDNN version: {format_cudnn_version(torch.backends.cudnn.version())}"
    )


def log_devices_info(filepath: Optional[Path] = None) -> None:
    """Logs high-level info about all available accelerator devices."""
    if not torch.cuda.is_available():
        return

    ncpus = os.cpu_count()
    num_devices = torch.cuda.device_count()
    log_lines = [f"CPU cores: {ncpus} CUDA devices: {num_devices}"]

    def _mem_to_gib(x):
        return round(float(x) / 1024**3, 2)

    for i in range(num_devices):
        device_name = torch.cuda.get_device_name(i)
        mem_free, mem_total = torch.cuda.mem_get_info(i)
        mem_allocated = torch.cuda.memory_allocated(i)
        mem_reserved = torch.cuda.memory_reserved(i)
        capability = torch.cuda.get_device_capability(i)
        log_lines.append(
            f"device({i})='{device_name}' "
            f"Capability: {capability} "
            f"Memory: [Total: {_mem_to_gib(mem_total)}GiB "
            f"Free: {_mem_to_gib(mem_free)}GiB "
            f"Allocated: {_mem_to_gib(mem_allocated)}GiB "
            f"Cached: {_mem_to_gib(mem_reserved)}GiB]"
        )

    all_text = "\n".join(log_lines)
    logger.info(all_text)

    if filepath:
        with filepath.open("w", encoding="utf-8") as f:
            f.write(all_text)


def log_peak_gpu_memory():
    """Log the peak GPU memory usage."""
    if torch.cuda.is_available():
        peak_memory = torch.cuda.max_memory_allocated() / 1024**3  # Convert to GB
        logger.info(f"Peak GPU memory usage: {peak_memory:.2f} GB")


def create_model_summary(model: Any) -> str:
    """Creates a model summary as a free-formed string."""
    lines = ["Model summary:", repr(model), ""]

    module_lines = [f"{name} ({type(layer)})" for name, layer in model.named_modules()]

    lines.append(f"Modules ({len(module_lines)}):")
    lines.extend(module_lines)
    lines.append("")

    # TODO: Consider whether to use `torchsummary` library here.
    # Caveat: it may require sample inputs/shapes, and other aux info.
    return "\n".join(lines)


def log_model_summary(model, filepath: Optional[Path] = None) -> None:
    """Logs a model summary."""
    model_summary = create_model_summary(model)
    logger.info(model_summary)

    if filepath:
        with filepath.open("w", encoding="utf-8") as f:
            f.write(model_summary)


def get_device_name() -> str:
    """Returns the name of the device, assuming all are identical."""
    device_name = "CPU"
    if torch.cuda.is_available():
        # Assume all devices are identical
        device_name = torch.cuda.get_device_name(0)
    elif torch.backends.mps.is_available():
        device_name = "MPS"
    return device_name


@dataclass(frozen=True)
class ModelParameterCount:
    all_params: int
    trainable_params: int
    embedding_params: int

    def __post_init__(self):
        """Ensure that the parameters are valid."""
        for name, value in [
            ("all_params", self.all_params),
            ("trainable_params", self.trainable_params),
            ("embedding_params", self.embedding_params),
        ]:
            if value < 0:
                raise ValueError(f"`{name}` ({value}) must be >= 0.")
            if self.trainable_params > self.all_params:
                raise ValueError(
                    f"`trainable_params` ({self.trainable_params}) cannot be "
                    f"greater than `all_params` ({self.all_params})."
                )
            if self.embedding_params > self.all_params:
                raise ValueError(
                    f"`embedding_params` ({self.embedding_params}) cannot be "
                    f"greater than `all_params` ({self.all_params})."
                )

    @property
    def trainable_params_percent(self) -> float:
        """Percentage of trainable parameters [0.0, 100.0]."""
        if self.all_params == 0:
            return 0.0
        return 100 * self.trainable_params / self.all_params

    @property
    def frozen_params_percent(self) -> float:
        """Percentage of frozen parameters [0.0, 100.0]."""
        return 100.0 - self.trainable_params_percent


def _get_parameter_names(
    model: torch.nn.Module, forbidden_layer_types: list[Any]
) -> list[str]:
    """Returns the names of the model parameters that are not inside a forbidden layer.

    Borrowed from
    https://github.com/huggingface/transformers/blob/main/src/transformers/trainer.py.
    """
    result = []
    for name, child in model.named_children():
        result += [
            f"{name}.{n}"
            for n in _get_parameter_names(child, forbidden_layer_types)
            if not isinstance(child, tuple(forbidden_layer_types))
        ]
    # Add model specific parameters (defined with nn.Parameter) since they are not in
    # any child.
    result += list(model._parameters.keys())
    return result


def count_model_parameters(model: torch.nn.Module) -> ModelParameterCount:
    """Creates a basic counter of the parameters in a neural model.

    Args:
        model: The torch-implemented neural network.

    Returns:
        ModelParameterCount: A ModelParameterCount for the underlying model.
    """
    trainable_params = 0
    all_params = 0
    embedding_params = 0
    embedding_layer_names = []
    for name, module in model.named_modules():
        if isinstance(module, torch.nn.Embedding):
            # Embedding layers appear in named_parameters with ".weight" at the end
            param_name = f"{name}.weight" if name else "weight"
            embedding_layer_names.append(param_name)

    for name, param in model.named_parameters():
        param_count = param.numel()
        all_params += param_count
        if param.requires_grad:
            trainable_params += param_count

        if name in embedding_layer_names:
            embedding_params += param_count

    return ModelParameterCount(
        all_params=all_params,
        trainable_params=trainable_params,
        embedding_params=embedding_params,
    )


def log_number_of_model_parameters(
    model: torch.nn.Module, use_icons: bool = True
) -> None:
    """Logs the number of parameters of the model.

    Args:
        model: The torch-implemented neural network.
        use_icons: Whether to display emojis/icons in the log output.
    """
    params = count_model_parameters(model)

    # Icons if enabled, else fallback to plain text
    total_label = "🔢 Total" if use_icons else "Total"
    embedding_label = "🔗 Embedding" if use_icons else "Embedding"
    trainable_label = "🎯 Trainable" if use_icons else "Trainable"
    frozen_label = "🔒 Frozen" if use_icons else "Frozen"

    n_space = 11 if use_icons else 9

    logger.info(
        f"\nModel Parameters Summary:\n"
        f"{total_label:<{n_space}} parameters: {params.all_params:,}\n"
        f"{embedding_label:<{n_space}} parameters: {params.embedding_params:,}\n"
        f"{trainable_label:<{n_space}} parameters: {params.trainable_params:,}\n"
        f"{frozen_label:<{n_space}} parameters: "
        f"{params.all_params - params.trainable_params:,} "
        f"({params.frozen_params_percent:.2f}%)\n"
    )


def get_torch_dtype(torch_dtype_str: str) -> torch.dtype:
    """Converts string dtype to torch.dtype."""
    torch_dtype_str = torch_dtype_str.lower()
    if torch_dtype_str in ["f64", "float64", "double"]:
        return torch.float64
    elif torch_dtype_str in ["f32", "float32", "float"]:
        return torch.float32
    elif torch_dtype_str in ["bf16", "bfloat16"]:
        return torch.bfloat16
    elif torch_dtype_str in ["f16", "float16", "half"]:
        return torch.float16
    elif torch_dtype_str in ["uint8"]:
        return torch.uint8
    else:
        raise ValueError(f"Unsupported torch dtype: {torch_dtype_str}")


def get_dtype_size_in_bytes(
    dtype: Union[str, torch.dtype, type[np.generic]],
) -> int:
    """Returns size of this dtype in bytes."""
    if isinstance(dtype, torch.dtype):
        return dtype.itemsize
    elif isinstance(dtype, str):
        if not dtype:
            raise ValueError("Empty string is not a valid dtype")
        try:
            # Try to parse using non-standard names like "f64"
            return get_torch_dtype(dtype).itemsize
        except ValueError:
            return np.dtype(dtype).itemsize

    return np.dtype(dtype).itemsize


def _estimate_item_size_in_bytes(item: Any) -> int:
    if isinstance(item, (int, float)):
        return 4
    elif isinstance(item, (np.ndarray, torch.Tensor)):
        num_elements = math.prod(item.shape)
        return num_elements * get_dtype_size_in_bytes(item.dtype)
    elif isinstance(item, list):
        return _estimate_sample_list_size_in_bytes(item)
    elif isinstance(item, str):
        return compute_utf8_len(item)
    elif isinstance(item, (str, bytes)):
        return len(item)

    return 0


def _estimate_sample_list_size_in_bytes(sample_list: list[Any]) -> int:
    num_elements = len(sample_list)
    if num_elements <= 0:
        return 0
    return sum(_estimate_item_size_in_bytes(item) for item in sample_list)


def estimate_sample_dict_size_in_bytes(sample: dict[str, Any]) -> int:
    """Estimates the approximate total number of bytes in a provided sample.

    Training sample is expected to be a dictionary, where a value is a list,
    tensor, or a numpy array.

    The function works in best effort mode i.e., 100% accuaracy isn't guaranteed.
    The implementation is slow, and shouldn't be called in performance-sensitive code.
    """
    result = 0
    for key, val in sample.items():
        result += compute_utf8_len(key)
        result += _estimate_item_size_in_bytes(val)
    return result


def coerce_model_to_dtype(model: torch.nn.Module, dtype: torch.dtype) -> None:
    """Coerces the model to the desired dtype.

    This is needed as a temporary workaround to support QLoRA FSDP training. See:
    https://github.com/huggingface/accelerate/issues/1620#issuecomment-2407102051
    """
    for name, module in model.named_modules():
        try:
            module.to(dtype)
        except Exception as e:
            logger.warning(
                f"Failed to coerce module {name} to dtype {dtype}. Error: {e}"
            )


T = TypeVar("T")


def convert_to_list_of_tensors(values: list[T]) -> list[torch.Tensor]:
    """Converts a list of array-like objects into alist of torch tensors."""
    if len(values) == 0:
        return []

    first_item = values[0]
    if isinstance(first_item, torch.Tensor):
        return [cast(torch.Tensor, item) for item in values]

    if isinstance(first_item, np.ndarray):
        return [torch.from_numpy(item) for item in values]
    elif isinstance(first_item, list):
        return [torch.from_numpy(np.asarray(item)) for item in values]

    raise ValueError(
        f"Unsupported element type: {type(first_item)}. "
        "Must be numpy array, torch tensor, or Python list."
    )


def pad_sequences_right_side(
    sequences: list[T], *, padding_value: float = 0
) -> torch.Tensor:
    """Pads a list of variable-length tensors to a single tensor.

    Appends `padding_value` to the right side of each sequence
    to expand to the longest length.

    Args:
        sequences: list of variable length sequences.
        padding_value: value for padded elements. Default: 0.

    Returns:
        A tensor with shape (B, L, ...), where B is a batch size (`len(sequences)`),
        L is the longest length (`max(len(sequences[i]))`)
    """
    return pad_sequences(sequences, padding_value=padding_value, padding_side="right")


def pad_sequences_left_side(
    sequences: list[T], *, padding_value: float = 0
) -> torch.Tensor:
    """Pads a list of variable-length tensors to a single tensor.

    Prepends `padding_value` to the left side of each sequence
    to expand to the longest length.

    Args:
        sequences: list of variable length sequences.
        padding_value: value for padded elements. Default: 0.

    Returns:
        A tensor with shape (B, L, ...), where B is a batch size (`len(sequences)`),
        L is the longest length (`max(len(sequences[i]))`)
    """
    return pad_sequences(sequences, padding_value=padding_value, padding_side="left")


def pad_sequences(
    sequences: list[T], *, padding_value: float = 0, padding_side: Optional[str] = None
) -> torch.Tensor:
    """Pads a list of variable-length tensors to a single tensor.

    Args:
        sequences: list of variable length sequences.
        padding_value: value for padded elements. Default: 0.
        padding_side: side to apply padding to. Valid values:  'right', 'left'.
            If unspecified (`None`), defaults to `right`.

    Returns:
        A tensor with shape (B, L, ...), where B is a batch size (`len(sequences)`),
        L is the longest length (`max(len(sequences[i]))`)
    """
    if not padding_side:
        padding_side = "right"

    if padding_side not in ("right", "left"):
        raise ValueError(
            f"Unsupported padding side: '{padding_side}'. "
            "Valid values: 'right', 'left'."
        )

    if len(sequences) == 0:
        raise ValueError("Empty list is not allowed.")
    tensor_sequences = convert_to_list_of_tensors(sequences)

    try:
        return torch.nn.utils.rnn.pad_sequence(
            tensor_sequences,
            batch_first=True,
            padding_value=padding_value,
            padding_side=padding_side,
        )
    except RuntimeError:
        logger.error(
            "Failed to pad sequences with the shapes: "
            + ", ".join([f"{t.shape}" for t in tensor_sequences])
        )
        raise


class _DimMinMaxSizes(NamedTuple):
    dim_index: int

    min_size: int
    max_size: int

    @property
    def has_variable_sizes(self) -> bool:
        return self.min_size != self.max_size


def _get_dims_min_max_size(tensors_list: list[torch.Tensor]) -> list[_DimMinMaxSizes]:
    num_tensors = len(tensors_list)
    if num_tensors <= 0:
        return []

    first_shape = tensors_list[0].shape

    min_dim_sizes = list(first_shape)
    max_dim_sizes = list(first_shape)
    num_dims = len(min_dim_sizes)

    for tensor_idx in range(num_tensors - 1):
        curr_shape = tensors_list[tensor_idx + 1].shape
        if num_dims != len(curr_shape):
            raise ValueError(
                "Tensors have different number of dimensions: "
                f"{num_dims} vs {len(curr_shape)}! "
                f"Shapes: {first_shape}, {curr_shape}"
            )
        for idx in range(num_dims):
            min_dim_sizes[idx] = min(min_dim_sizes[idx], curr_shape[idx])
            max_dim_sizes[idx] = max(max_dim_sizes[idx], curr_shape[idx])

    return [
        _DimMinMaxSizes(
            dim_index=idx, min_size=min_dim_sizes[idx], max_size=max_dim_sizes[idx]
        )
        for idx in range(num_dims)
    ]


def _format_dims_min_max_sizes(dim_sizes: list[_DimMinMaxSizes]) -> str:
    result: list[str] = [""] * len(dim_sizes)
    for idx, item in enumerate(dim_sizes):
        result[idx] = (
            f"{item.min_size}...{item.max_size}"
            if item.has_variable_sizes
            else f"{item.min_size}"
        )
    return "[" + ", ".join(result) + "]"


def _pad_to_max_dim_and_stack_impl(
    tensors_list: list[torch.Tensor],
    *,
    max_variable_sized_dims: int,
    padding_value: float,
    pad_on_left_side: bool,
) -> torch.Tensor:
    num_tensors = len(tensors_list)
    if num_tensors == 0:
        raise ValueError("Empty list of tensors is not allowed.")
    dim_sizes: list[_DimMinMaxSizes] = _get_dims_min_max_size(tensors_list)
    num_variable_size_dims = sum(
        (1 if item.has_variable_sizes else 0) for item in dim_sizes
    )
    if (
        max_variable_sized_dims >= 0
        and num_variable_size_dims > max_variable_sized_dims
    ):
        raise ValueError(
            "Too many dimensions with variable size. "
            f"Got: {num_variable_size_dims} variable size dimensions. "
            f"Maximum allowed: {max_variable_sized_dims}. "
            f"Dimension sizes: {_format_dims_min_max_sizes(dim_sizes)}."
        )

    if num_variable_size_dims == 0:
        # No need to pad anything, just `stack()`.
        return torch.stack(tensors_list)
    elif num_variable_size_dims == 1 and dim_sizes[0].has_variable_sizes:
        # Use pad_sequences provided by PyTorch, which should be equivalent
        # for the common case.
        if pad_on_left_side:
            return pad_sequences_left_side(tensors_list, padding_value=padding_value)
        return pad_sequences_right_side(tensors_list, padding_value=padding_value)

    max_dim_sizes = [item.max_size for item in dim_sizes]
    result_shape = torch.Size([num_tensors] + max_dim_sizes)

    result = torch.full(
        result_shape,
        padding_value,
        dtype=tensors_list[0].dtype,
        device=tensors_list[0].device,
    )

    max_dim_sizes = torch.Size(max_dim_sizes)
    for i, input_tensor in enumerate(tensors_list):
        target = result.select(0, i)
        if max_dim_sizes == input_tensor.shape:
            target[...] = input_tensor
            continue

        target_view = target[...]

        for dim_idx, curr_size in enumerate(input_tensor.shape):
            max_size = max_dim_sizes[dim_idx]
            if curr_size < max_size:
                start_idx = (max_size - curr_size) if pad_on_left_side else 0
                target_view = target_view.narrow(
                    dim_idx, start=start_idx, length=curr_size
                )

        assert target_view.shape == input_tensor.shape
        target_view[...] = input_tensor

    return result


def pad_to_max_dim_and_stack(
    tensors_list: list[T],
    *,
    max_variable_sized_dims: int = -1,
    padding_value: float = 0,
    padding_side: Optional[str] = None,
) -> torch.Tensor:
    """Stacks variable-length tensors to a single tensor with dimension expansion.

    Some examples:
    1) Two tensors with shapes [24,8], [32,8] are combined to [2,32,8].
    2) Two tensors with shapes [24,1,8], [32,4,8] are combined to [2,32,4,8].
    3) Three tensors with shapes [7,3,5],[8,2,6],[9,1,7] are combined to [3,9,3,7].

    For 1D input tensors, the function is equivalent to `pad_sequences()`.

    If all tensors have the same shape and no padding is required, then the function
    is equivalent to `torch.stack()`.

    Args:
        tensors_list: list of tensors with potentially .
        max_variable_sized_dims: Maximum number of variable-sized dimensions.
            Negative values mean `Unlimited`.
            If you know that your tensors have a pre-defined number `N` of
            variable-sized dimensions (e.g., 1 for `sequence_length`) then
            it's a good idea to set this parameter to catch abnormal inputs
            (`ValueError` will be raised in such cases).
        padding_value: value for padded elements. Default: 0.
        padding_side: side to apply padding to. Valid values:  'right', 'left'.
            If unspecified (`None`), defaults to `right`.

    Returns:
        A tensor with shape (B, L, ...), where B is a batch size (`len(sequences)`),
        L is the longest length (`max(len(sequences[i]))`)
    """
    pad_on_left_side: bool = False
    if not padding_side or padding_side == "right":
        pad_on_left_side = False
    elif padding_side == "left":
        pad_on_left_side = True
    else:
        raise ValueError(
            f"Unsupported padding side: '{padding_side}'. "
            "Valid values: 'right', 'left'."
        )

    input_tensors = convert_to_list_of_tensors(tensors_list)

    try:
        return _pad_to_max_dim_and_stack_impl(
            input_tensors,
            max_variable_sized_dims=max_variable_sized_dims,
            padding_value=padding_value,
            pad_on_left_side=pad_on_left_side,
        )
    except RuntimeError:
        logger.error(
            "Failed to pad and stack tensors with the shapes: "
            + ", ".join([f"{t.shape}" for t in input_tensors])
        )
        raise


def create_ones_like(
    values: T,
) -> T:
    """Converts an array-like object into an object of the same type filled with 1-s.

    Supports nested lists, in which case all elements must be of the same type.
    """
    if isinstance(values, torch.Tensor):
        return torch.ones_like(values)
    elif isinstance(values, np.ndarray):
        return np.ones_like(values)
    elif not isinstance(values, list):
        raise ValueError(
            f"Unsupported type: {type(values)}. "
            "Must be numpy array, torch tensor, or Python list."
        )

    if len(values) == 0:
        return cast(T, [])

    first_item = values[0]
    if isinstance(first_item, (int, float)):
        result = list(np.ones_like(values))
    else:
        # Nested list
        first_item_type = type(first_item)
        result = []
        for idx, item in enumerate(values):
            if idx > 0 and not isinstance(item, first_item_type):
                raise ValueError(
                    "Sequence contains elements of different types: "
                    f"{first_item_type} and {type(item)}."
                )
            result.append(create_ones_like(item))

    return cast(T, result)


def get_first_dim_len(x: Any) -> int:
    """Returns length of the first dimension."""
    if isinstance(x, (torch.Tensor, np.ndarray)):
        return int(x.shape[0])
    elif isinstance(x, list):
        return len(x)

    raise ValueError(
        f"Unsupported type: {type(x)}. "
        "Must be numpy array, torch tensor, or Python list."
    )


def get_shape_as_list(x: Any) -> list[int]:
    """Returns shape of an object (tensor or numpy array) as Python list."""
    if isinstance(x, (torch.Tensor, np.ndarray)):
        return list(x.shape)

    raise ValueError(f"Unsupported type: {type(x)}. Must be numpy array, torch tensor.")


class _FreezeModelLayer:
    def __init__(self, name: str, freeze_it: bool):
        self.name: str = name
        self.freeze_it: bool = freeze_it
        self.children: list[_FreezeModelLayer] = []


def _freeze_model_layers_impl(
    module: torch.nn.Module, freeze_layers: list[_FreezeModelLayer], parent_path: str
) -> int:
    result: int = 0
    for model_layer in freeze_layers:
        full_layer_path = (
            (parent_path + "." + model_layer.name) if parent_path else model_layer.name
        )
        if hasattr(module, model_layer.name):
            child_module = getattr(module, model_layer.name)
            if model_layer.freeze_it:
                logger.info(f"Freezing layer '{full_layer_path}'...")
                for param in child_module.parameters(recurse=True):
                    param.requires_grad_(False)
                result += 1
            elif len(model_layer.children) > 0:
                result += _freeze_model_layers_impl(
                    child_module, model_layer.children, full_layer_path
                )
        else:
            logger.warning(f"Layer '{full_layer_path}' not found in model.")

    return result


def _group_freeze_model_layers(freeze_layers: list[str]) -> list[_FreezeModelLayer]:
    dummy_root: _FreezeModelLayer = _FreezeModelLayer(name="", freeze_it=False)

    # Build a tree of nested layers.
    for layer_name in freeze_layers:
        layer: _FreezeModelLayer = dummy_root
        all_parts = list(layer_name.split("."))
        for idx, curr_part in enumerate(all_parts):
            next_layer = next((x for x in layer.children if x.name == curr_part), None)
            # If it's the last part, let's freeze this layer.
            freeze_it = idx + 1 >= len(all_parts)
            if next_layer is None:
                next_layer = _FreezeModelLayer(name=curr_part, freeze_it=freeze_it)
                layer.children.append(next_layer)
            elif freeze_it:
                next_layer.freeze_it = True
            layer = next_layer
    return dummy_root.children


def freeze_model_layers(model: torch.nn.Module, freeze_layers: list[str]) -> int:
    """Recursively freezes model layers.

    Args:
        model: A model to freeze layers in.
        freeze_layers: A list of layer names to freeze.
            Nested layers can be specified using a dot ('.') separator.
            For example, "visual.child.grandchild".
            Layer names not found in the model are ignored.

    Returns:
        The total number of layers successfully frozen.
    """
    root_freeze_layers = _group_freeze_model_layers(freeze_layers)

    return _freeze_model_layers_impl(model, root_freeze_layers, "")



from unittest.mock import MagicMock, patch

import pytest
import torch
from torch.utils.data import DataLoader
from torchdata.stateful_dataloader import StatefulDataLoader

from oumi.core.configs import TelemetryParams, TrainingParams
from oumi.core.configs.params.fsdp_params import FSDPParams
from oumi.core.trainers.oumi_trainer import Trainer
from oumi.models import MLPEncoder
from tests.markers import requires_gpus


#
# Fixtures
#
@pytest.fixture
def model():
    return MLPEncoder()


@pytest.fixture
def mock_model():
    return MagicMock(spec=torch.nn.Module)


@pytest.fixture
def mock_dataset():
    dataset = MagicMock()
    dataset.__len__ = MagicMock(return_value=10)
    dataset.state_dict = None
    return dataset


@pytest.fixture
def mock_stateful_dataloader():
    sample_batch = {
        "input_ids": torch.randint(0, 1000, (4, 768)),
        "attention_mask": torch.ones(4, 768),
        "labels": torch.randint(0, 1000, (4, 768)),
    }

    mock_loader = MagicMock(spec=StatefulDataLoader)
    mock_loader.__iter__.return_value = iter([sample_batch] * 10)
    mock_loader.__len__.return_value = 10

    return mock_loader


@pytest.fixture
def mock_dataloader():
    sample_batch = {
        "input_ids": torch.randint(0, 1000, (4, 768)),
        "attention_mask": torch.ones(4, 768),
        "labels": torch.randint(0, 1000, (4, 768)),
    }

    mock_loader = MagicMock(spec=DataLoader)
    mock_loader.__iter__.return_value = iter([sample_batch] * 10)
    mock_loader.__len__.return_value = 10

    return mock_loader


@pytest.fixture
def mock_optimizer():
    return MagicMock(spec=torch.optim.Optimizer)


@pytest.fixture
def mock_params():
    args = MagicMock(spec=TrainingParams)
    args.adam_beta1 = 0.9
    args.adam_beta2 = 0.999
    args.adam_epsilon = 1e-8
    args.compile = False
    args.dataloader_num_workers = 0
    args.dataloader_prefetch_factor = 2
    args.enable_gradient_checkpointing = False
    args.enable_tensorboard = False
    args.enable_wandb = False
    args.enable_mlflow = False
    args.eval_steps = 50
    args.eval_strategy = "steps"
    args.gradient_accumulation_steps = 1
    args.learning_rate = 0.001
    args.learning_rate = 0.001
    args.logging_steps = 10
    args.lr_scheduler_kwargs = {}
    args.lr_scheduler_type = "linear"
    args.max_steps = 100
    args.num_train_epochs = 3
    args.optimizer = "adamw"
    args.output_dir = "/tmp/test_output"
    args.per_device_eval_batch_size = 8
    args.per_device_train_batch_size = 8
    args.save_epoch = True
    args.save_steps = 50
    args.warmup_ratio = None
    args.warmup_steps = 0
    args.weight_decay = 0.01
    return args


@pytest.fixture
def trainer(model, mock_tokenizer, mock_params, mock_dataset):
    return Trainer(
        model=model,
        processing_class=mock_tokenizer,
        processor=None,
        args=mock_params,
        train_dataset=mock_dataset,
        eval_dataset=mock_dataset,
    )


#
# Tests
#
def test_trainer_initialization(
    trainer, model, mock_tokenizer, mock_params, mock_dataset
):
    assert trainer.model == model
    assert trainer.processing_class == mock_tokenizer
    assert trainer.params == mock_params
    assert trainer.train_dataset == mock_dataset
    assert trainer.eval_dataset == mock_dataset
    assert isinstance(trainer.optimizer, torch.optim.AdamW)
    assert isinstance(trainer.train_dataloader, StatefulDataLoader)
    assert isinstance(trainer.eval_dataloader, DataLoader)
    assert trainer.state.epoch == 0
    assert trainer.state.global_step == 0


def test_get_total_training_steps(trainer):
    if trainer.params.max_steps is not None:
        assert trainer._estimate_total_training_steps() == trainer.params.max_steps


@patch("oumi.core.distributed.is_world_process_zero", return_value=True)
def test_train_with_num_epochs(mock_is_world_process_zero, trainer):
    trainer._train_epoch = MagicMock()
    trainer.save_state = MagicMock()
    trainer.evaluate = MagicMock()

    trainer.params.eval_strategy = "epoch"
    trainer.params.max_steps = -1  # Use `num_train_epochs`

    trainer.train()

    assert trainer._train_epoch.call_count == trainer.params.num_train_epochs
    assert trainer.save_state.call_count == trainer.params.num_train_epochs
    assert trainer.evaluate.call_count == trainer.params.num_train_epochs


@patch("oumi.core.distributed.is_world_process_zero", return_value=True)
def test_train_with_max_steps(mock_is_world_process_zero, trainer):
    steps_per_epoch: int = 5

    def _increment_global_step():
        trainer.state.global_step += steps_per_epoch

    trainer._train_epoch = MagicMock()
    trainer._train_epoch.side_effect = lambda pbar: _increment_global_step()
    trainer.save_state = MagicMock()
    trainer.evaluate = MagicMock()

    trainer.params.eval_strategy = "epoch"
    trainer.params.max_steps = 100
    expected_epochs = trainer.params.max_steps // steps_per_epoch

    trainer.train()

    assert trainer._train_epoch.call_count == expected_epochs
    assert trainer.save_state.call_count == expected_epochs
    assert trainer.evaluate.call_count == expected_epochs


def test_train_epoch(trainer, mock_stateful_dataloader, tmp_path):
    output_dir = tmp_path / "model_output"
    output_dir.mkdir()

    trainer._process_callbacks = MagicMock()
    trainer.telemetry.timer = MagicMock()
    trainer.model.forward = MagicMock(
        return_value={"loss": torch.tensor(0.5), "logits": torch.tensor([1.0, 2.0])}
    )
    trainer.train_dataloader = mock_stateful_dataloader
    trainer.params.telemetry = MagicMock(spec=TelemetryParams)
    trainer.params.telemetry.collect_telemetry_for_all_ranks = False
    trainer.params.telemetry_dir = MagicMock(return_value=(output_dir / "telemetry"))
    trainer.scaler.scale = MagicMock(return_value=MagicMock())
    trainer.scaler.step = MagicMock()
    trainer.scaler.update = MagicMock()

    progress_bar = MagicMock()
    trainer._train_epoch(progress_bar)

    assert trainer._process_callbacks.call_count > 0
    assert trainer.telemetry.timer.call_count > 0
    assert trainer.model.forward.call_count > 0
    assert trainer.scaler.scale.call_count > 0
    assert trainer.scaler.step.call_count > 0
    assert trainer.scaler.update.call_count > 0


def test_evaluate(trainer, mock_dataloader):
    trainer.model.eval = MagicMock()
    trainer.model.forward = MagicMock(return_value=MagicMock(loss=torch.tensor(0.5)))
    trainer.eval_dataloader = mock_dataloader

    results = trainer.evaluate()

    assert "val/loss" in results
    assert "val/perplexity" in results
    assert trainer.model.eval.call_count == 1
    assert trainer.model.forward.call_count > 0


@pytest.fixture
def mock_dcp_save():
    with patch("torch.distributed.checkpoint.save") as mock_save:
        yield mock_save


@pytest.fixture
def mock_dcp_load():
    with patch("torch.distributed.checkpoint.load") as mock_load:
        yield mock_load


@pytest.mark.parametrize("is_using_fsdp", [True, False])
def test_save_and_load_model(
    trainer,
    mock_model,
    mock_optimizer,
    mock_stateful_dataloader,
    tmp_path,
    mock_dcp_save,
    mock_dcp_load,
    is_using_fsdp,
):
    output_dir = tmp_path / "model_output"
    output_dir.mkdir()
    telemetry_dir = output_dir / "telemetry"
    telemetry_dir.mkdir()

    trainer.fsdp_params = FSDPParams(enable_fsdp=is_using_fsdp)
    trainer.is_using_fsdp = is_using_fsdp
    trainer.train_dataloader = mock_stateful_dataloader
    trainer.params.output_dir = str(output_dir)
    trainer.params.telemetry = MagicMock(spec=TelemetryParams)
    trainer.params.telemetry.collect_telemetry_for_all_ranks = False
    trainer.params.telemetry_dir = telemetry_dir

    trainer.train_dataloader.state_dict = MagicMock(
        return_value={"dataloader_key": torch.tensor(3)}
    )
    trainer.state.epoch = 1
    trainer.state.global_step = 50

    with patch("oumi.core.trainers.oumi_trainer.get_state_dict") as mock_get_state_dict:
        mock_get_state_dict.return_value = ({"model": "state"}, {"optimizer": "state"})

        trainer.save_state()

        mock_dcp_save.assert_called()

        assert (output_dir / "dataloader.pt").exists()
        assert (output_dir / "trainer_state.json").exists()
        assert (telemetry_dir / "telemetry_rank0000.json").exists()

        # Folder are created by DCP, but since it's a mock, we need to create
        # the folder manually.
        (output_dir / "model").mkdir()
        (output_dir / "optimizer").mkdir()

        # Test load
        trainer._load_from_checkpoint(str(output_dir))
        mock_dcp_load.assert_called()
        assert trainer.train_dataloader.load_state_dict.called
        assert trainer.state.epoch == 1
        assert trainer.state.global_step == 50


def test_get_train_dataloader(trainer):
    dataloader = trainer._get_train_dataloader()
    assert isinstance(dataloader, StatefulDataLoader)
    assert dataloader.batch_size == trainer.params.per_device_train_batch_size


def test_get_eval_dataloader(trainer):
    dataloader = trainer._get_eval_dataloader()
    assert isinstance(dataloader, DataLoader)
    assert dataloader.batch_size == trainer.params.per_device_eval_batch_size


def test_process_callbacks(trainer):
    mock_callback = MagicMock()
    mock_callback.on_log = MagicMock()
    trainer.callbacks = [mock_callback]

    logs = trainer._process_callbacks("on_log")

    assert mock_callback.on_log.called
    assert isinstance(logs, dict)


@requires_gpus()
def test_cuda_initialization(model, mock_tokenizer, mock_params, mock_dataset):
    assert next(model.parameters()).is_cpu
    trainer = Trainer(
        model=model,
        processing_class=mock_tokenizer,
        processor=None,
        args=mock_params,
        train_dataset=mock_dataset,
        eval_dataset=None,
    )
    assert next(model.parameters()).is_cuda, "Model should be on CUDA"
    assert trainer.device.startswith("cuda"), "Device should be CUDA"


@pytest.mark.skipif(not torch.backends.mps.is_available(), reason="MPS not available")
def test_mps_initialization(model, mock_tokenizer, mock_params, mock_dataset):
    assert next(model.parameters()).is_cpu, "Model should initially be on CPU"
    trainer = Trainer(
        model=model,
        processing_class=mock_tokenizer,
        processor=None,
        args=mock_params,
        train_dataset=mock_dataset,
        eval_dataset=None,
    )
    assert next(model.parameters()).is_mps, "Model should be on MPS"
    assert trainer.device == "mps", "Device should be MPS"


#
# MLflow Tests
#
@pytest.fixture
def mock_mlflow():
    """Mock MLflow module for testing."""
    with patch("oumi.core.trainers.oumi_trainer.mlflow") as mock_mlflow:
        # Mock active_run to return None by default (no active run)
        mock_mlflow.active_run.return_value = None
        # Mock start_run to return a mock run
        mock_run = MagicMock()
        mock_run.info.run_id = "test-run-id"
        mock_mlflow.start_run.return_value = mock_run
        # Mock log_metrics
        mock_mlflow.log_metrics = MagicMock()
        # Mock end_run
        mock_mlflow.end_run = MagicMock()
        yield mock_mlflow


@pytest.fixture
def mlflow_params():
    """Training params with MLflow enabled."""
    args = MagicMock(spec=TrainingParams)
    args.adam_beta1 = 0.9
    args.adam_beta2 = 0.999
    args.adam_epsilon = 1e-8
    args.compile = False
    args.dataloader_num_workers = 0
    args.dataloader_prefetch_factor = 2
    args.enable_gradient_checkpointing = False
    args.enable_tensorboard = False
    args.enable_wandb = False
    args.enable_mlflow = True  # Enable MLflow
    args.eval_steps = 50
    args.eval_strategy = "steps"
    args.gradient_accumulation_steps = 1
    args.learning_rate = 0.001
    args.logging_steps = 10
    args.lr_scheduler_kwargs = {}
    args.lr_scheduler_type = "linear"
    args.max_steps = 100
    args.num_train_epochs = 3
    args.optimizer = "adamw"
    args.output_dir = "/tmp/test_output"
    args.per_device_eval_batch_size = 8
    args.per_device_train_batch_size = 8
    args.save_epoch = True
    args.save_steps = 50
    args.warmup_ratio = None
    args.warmup_steps = 0
    args.weight_decay = 0.01
    args.run_name = "test-run"
    return args


def test_mlflow_initialization_with_no_active_run(
    model, mock_tokenizer, mlflow_params, mock_dataset, mock_mlflow
):
    with patch("oumi.core.distributed.is_world_process_zero", return_value=True):
        trainer = Trainer(
            model=model,
            processing_class=mock_tokenizer,
            processor=None,
            args=mlflow_params,
            train_dataset=mock_dataset,
            eval_dataset=mock_dataset,
        )

        mock_mlflow.start_run.assert_called_once_with(run_name="test-run")
        assert trainer._mlflow_oumi_managed_run is True


def test_mlflow_initialization_with_active_run(
    model, mock_tokenizer, mlflow_params, mock_dataset, mock_mlflow
):
    mock_mlflow.active_run.return_value = MagicMock()

    with patch("oumi.core.distributed.is_world_process_zero", return_value=True):
        trainer = Trainer(
            model=model,
            processing_class=mock_tokenizer,
            processor=None,
            args=mlflow_params,
            train_dataset=mock_dataset,
            eval_dataset=mock_dataset,
        )

        # Check that MLflow run was NOT started (user manages it)
        mock_mlflow.start_run.assert_not_called()
        assert trainer._mlflow_oumi_managed_run is False


def test_mlflow_log_metrics(trainer, mock_mlflow):
    trainer.params.enable_mlflow = True
    trainer.state.global_step = 10

    metrics = {"loss": 0.5, "accuracy": 0.8}

    with patch("oumi.core.distributed.is_world_process_zero", return_value=True):
        trainer.log_metrics(metrics, step=10)

        mock_mlflow.log_metrics.assert_called_once_with(metrics, step=10)



from dataclasses import dataclass
from typing import Any

import pytest

from oumi.core.configs.params.base_params import BaseParams


@dataclass
class NestedParams(BaseParams):
    simple_param: Any
    list_param: list
    dict_param: dict


@dataclass
class PositiveValueParams(BaseParams):
    value: Any

    def __finalize_and_validate__(self) -> None:
        if self.value <= 0:
            raise ValueError(f"Value must be positive, got {self.value}")


@dataclass
class ValueParams(BaseParams):
    value: Any


#
# Tests
#
def test_simple_params():
    simple = PositiveValueParams(42)
    simple.finalize_and_validate()  # Should not raise any exception


def test_nested_params():
    simple1 = PositiveValueParams(1)
    simple2 = PositiveValueParams(2)
    nested = NestedParams(
        simple_param=simple1,
        list_param=[simple2, "not a param"],
        dict_param={"a": PositiveValueParams(3), "b": "not a param"},
    )
    nested.finalize_and_validate()  # Should not raise any exception


def test_validation_call_count():
    @dataclass
    class CounterParams(BaseParams):
        call_count = 0
        a: Any = None
        b: Any = None
        c: Any = None

        def __finalize_and_validate__(self):
            CounterParams.call_count += 1

    root = CounterParams(
        a=CounterParams(),
        b=[CounterParams(), "not a param"],
        c={
            "x": CounterParams(),
            "y": "not a param",
        },
    )

    assert CounterParams.call_count == 0  # before validation
    root.finalize_and_validate()
    assert CounterParams.call_count == 4  # root + 3 unique nested CounterParams


def test_cyclic_reference():
    a = ValueParams(1)
    b = ValueParams(2)
    a.value = b
    b.value = a

    a.finalize_and_validate()  # Should not cause infinite recursion


def test_nested_params_with_failing_child():
    # Should not fail post init
    failing_param = PositiveValueParams(-5)

    # Should fail after validation
    with pytest.raises(expected_exception=ValueError, match="Value must be positive"):
        failing_param.finalize_and_validate()

    # Should fail after validation of parent
    nested = NestedParams(
        simple_param=PositiveValueParams(42),
        list_param=[
            PositiveValueParams(1),
            failing_param,
        ],
        dict_param={"a": PositiveValueParams(3), "b": PositiveValueParams(4)},
    )

    with pytest.raises(expected_exception=ValueError, match="Value must be positive"):
        nested.finalize_and_validate()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("allenai/c4")
class C4Dataset(BasePretrainingDataset):
    """A dataset for pretraining on the Colossal Clean Crawled Corpus (C4).

    The C4 dataset is based on the Common Crawl dataset and is available in
    multiple variants: 'en', 'en.noclean', 'en.noblocklist', 'realnewslike',
    and 'multilingual' (mC4). It is intended for pretraining language models
    and word representations.

    For more details and download instructions, visit:
    https://huggingface.co/datasets/allenai/c4

    References:
        Paper: https://arxiv.org/abs/1910.10683

    Data Fields:
        - url: URL of the source as a string
        - text: Text content as a string
        - timestamp: Timestamp as a string

    Dataset Variants:
        - en: 305GB
        - en.noclean: 2.3TB
        - en.noblocklist: 380GB
        - realnewslike: 15GB
        - multilingual (mC4): 9.7TB (108 subsets, one per language)

    The dataset is released under the ODC-BY license and is subject to the
    Common Crawl terms of use.
    """

    default_dataset = "allenai/c4"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
from typing import Optional, Union

import pydantic
from typing_extensions import Self

from oumi.core.configs.params.judge_params import (
    JudgeOutputType,
    JudgeResponseFormat,
)
from oumi.core.inference import BaseInferenceEngine
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.utils.placeholders import resolve_placeholders


class JudgeOutputField(pydantic.BaseModel):
    """Represents a single output field that a judge can produce.

    Attributes:
        field_key: The key/name for this field in the judge's output
        field_type: The data type expected for this field's value
        field_scores: Optional mapping from categorical values to numeric scores
    """

    field_key: str
    field_type: JudgeOutputType
    field_scores: Optional[dict[str, float]]

    def get_typed_value(self, raw_value: str) -> Optional[Union[float, int, str, bool]]:
        """Convert the field's raw string value to the appropriate type.

        Args:
            raw_value: The raw string value from the judge's output

        Returns:
            The typed value, or None if conversion fails

        Raises:
            ValueError: If the field_type is not supported
        """
        if self.field_type == JudgeOutputType.BOOL:
            from oumi.utils.str_utils import try_str_to_bool

            return try_str_to_bool(raw_value)

        elif self.field_type == JudgeOutputType.INT:
            try:
                return int(raw_value)
            except ValueError:
                return None

        elif self.field_type == JudgeOutputType.FLOAT:
            try:
                return float(raw_value)
            except ValueError:
                return None

        elif self.field_type == JudgeOutputType.ENUM:
            if not self.field_scores or not isinstance(self.field_scores, dict):
                raise ValueError(
                    "ENUM type requires field_scores to map values to scores."
                )
            # Only return the raw value if it exists in the scores mapping
            return raw_value if raw_value in self.field_scores else None

        elif self.field_type == JudgeOutputType.TEXT:
            return raw_value

        else:
            raise ValueError(
                f"Unsupported field type: {self.field_type}. "
                "Supported types are: BOOL, INT, FLOAT, ENUM, TEXT."
            )


class JudgeOutput(pydantic.BaseModel):
    """Represents the output from a judge evaluation.

    Attributes:
        raw_output: The original unprocessed output from the judge
        parsed_output: Structured data (fields & their values) extracted from raw output
        output_fields: List of expected output fields for this judge
        field_values: Typed values for each expected output field
        field_scores: Numeric scores for each expected output field (if applicable)
        response_format: Format used for generating output (XML, JSON, or RAW)
    """

    raw_output: str
    parsed_output: dict[str, str] = {}
    output_fields: Optional[list[JudgeOutputField]] = None
    field_values: dict[str, Optional[Union[float, int, str, bool]]] = {}
    field_scores: dict[str, Optional[float]] = {}
    response_format: Optional[JudgeResponseFormat] = None

    @classmethod
    def from_raw_output(
        cls,
        raw_output: str,
        response_format: JudgeResponseFormat,
        output_fields: list[JudgeOutputField],
    ) -> Self:
        """Generate a structured judge output from a raw model output."""
        field_values = {}
        field_scores = {}

        # Parse the judge's response based on the expected format
        if response_format == JudgeResponseFormat.XML:
            parsed_output = cls._parse_xml_output(raw_output)
        elif response_format == JudgeResponseFormat.JSON:
            parsed_output = cls._parse_json_output(raw_output)
        else:  # JudgeResponseFormat.RAW
            parsed_output = {}

        # Process each expected output field
        for field in output_fields:
            if field.field_key not in parsed_output:
                field_values[field.field_key] = None
                field_scores[field.field_key] = None
                continue

            # Extract and clean the raw value
            raw_value = parsed_output[field.field_key].strip()

            # Convert to the appropriate type
            typed_value = field.get_typed_value(raw_value)
            field_values[field.field_key] = typed_value

            # Extract numeric score if field has score mapping
            if field.field_scores:
                field_scores[field.field_key] = field.field_scores.get(raw_value)
            elif field.field_type == JudgeOutputType.BOOL:
                # For boolean fields, scores can be inferred
                field_scores[field.field_key] = 1.0 if typed_value else 0.0
            else:
                field_scores[field.field_key] = None

        return cls(
            raw_output=raw_output,
            parsed_output=parsed_output,
            field_values=field_values,
            field_scores=field_scores,
            response_format=response_format,
            output_fields=output_fields,
        )

    @classmethod
    def _parse_xml_output(cls, xml_output: Optional[str]) -> dict[str, str]:
        """Parses an XML judge output."""
        if not xml_output:
            return {}

        # Regex pattern to match XML-like tags and their content
        # Captures the tag name in group 1 and the content between tags in group 2
        # For example, "<label>True</label>" would match as ("label", "True")
        pattern = r"<(\w+)>(.*?)</\1>"
        matches = re.findall(pattern, xml_output, re.DOTALL)

        return {field_name: field_value.strip() for field_name, field_value in matches}

    # TODO: Consider leveraging structured-outputs for better JSON parsing
    # https://oumi.ai/docs/en/latest/user_guides/infer/common_workflows.html#structured-outputs
    @classmethod
    def _parse_json_output(cls, json_output: Optional[str]) -> dict[str, str]:
        """Parse judgment data from JSON format.

        Args:
            json_output: Raw JSON string from the judge

        Returns:
            Dictionary of field names to values, empty dict if parsing fails
        """
        if not json_output:
            return {}

        # Remove any API formatting
        if json_output.startswith("```json"):
            json_output = json_output[len("```json") :].lstrip()
        if json_output.endswith("```"):
            json_output = json_output[:-3].rstrip()

        try:
            parsed = json.loads(json_output)
            # Ensure all values are strings for consistent processing
            return {k: str(v) for k, v in parsed.items()}
        except json.JSONDecodeError:
            return {}

    def generate_raw_output(self, field_values: dict[str, str]) -> str:
        """Generate raw output string from field values in the specified format.

        Args:
            field_values: Dictionary mapping field keys to their string values.
                            Must contain values for all required output fields.

        Returns:
            Formatted raw output string ready for use as assistant response.

        Raises:
            ValueError: If required output fields are missing from field_values,
                            if response_format/output_fields are not set, or if
                            response_format is not supported.
        """
        if not self.response_format:
            raise ValueError("response_format must be set before generating output")
        if not self.output_fields:
            raise ValueError("output_fields must be set before generating output")

        # Extract required field keys from output_fields
        required_field_keys = {field.field_key for field in self.output_fields}

        # Validate that all required fields are provided
        provided_keys = set(field_values.keys())
        if missing_keys := required_field_keys - provided_keys:
            raise ValueError(
                f"Missing values for required output fields: {sorted(missing_keys)}. "
                f"Required: {sorted(required_field_keys)}, "
                f"Provided: {sorted(provided_keys)}"
            )

        # Filter field_values to only include required output fields
        filtered_field_values = {
            key: value
            for key, value in field_values.items()
            if key in required_field_keys
        }

        if self.response_format == JudgeResponseFormat.XML:
            return self._generate_xml_output(filtered_field_values)
        elif self.response_format == JudgeResponseFormat.JSON:
            return self._generate_json_output(filtered_field_values)
        elif self.response_format == JudgeResponseFormat.RAW:
            # For RAW format, concatenate values in the order of output_fields
            ordered_values = [
                filtered_field_values[field.field_key] for field in self.output_fields
            ]
            return "\n".join(ordered_values)
        else:
            raise ValueError(f"Unsupported response format: {self.response_format}")

    @classmethod
    def _generate_xml_output(cls, field_values: dict[str, str]) -> str:
        """Generate XML formatted output from field values.

        Args:
            field_values: Dictionary mapping field keys to their string values.

        Returns:
            XML formatted string with each field as a tag.
        """
        xml_parts = [f"<{key}>{value}</{key}>" for key, value in field_values.items()]
        return "\n".join(xml_parts)

    @classmethod
    def _generate_json_output(cls, field_values: dict[str, str]) -> str:
        """Generate JSON formatted output from field values.

        Args:
            field_values: Dictionary mapping field keys to their string values.

        Returns:
            JSON formatted string.
        """
        return json.dumps(field_values, indent=2)

    def to_json(self) -> str:
        """Convert the JudgeOutput to a JSON string.

        Returns:
            JSON string representation of the JudgeOutput data.
        """
        return json.dumps(self.model_dump())


class BaseJudge:
    """Base class for implementing judges that evaluate model outputs.

    A judge takes structured inputs, formats them using a prompt template,
    runs inference to get judgments, and parses the results into structured outputs.
    """

    def __init__(
        self,
        prompt_template: str,
        prompt_template_placeholders: Optional[set[str]],
        system_instruction: Optional[str],
        example_field_values: list[dict[str, str]],
        response_format: JudgeResponseFormat,
        output_fields: list[JudgeOutputField],
        inference_engine: BaseInferenceEngine,
    ):
        """Initialize the judge.

        Args:
            prompt_template: Template string with placeholders for input data
            prompt_template_placeholders: Set of expected placeholders in template
            system_instruction: Optional system message to guide judge behavior
            example_field_values: List of field value dicts for few-shot learning
            response_format: Expected format of judge responses (XML, JSON, or RAW)
            output_fields: List of fields expected in judge outputs
            inference_engine: Engine for running model inference
        """
        self.prompt_template = prompt_template
        self.prompt_template_placeholders = prompt_template_placeholders
        self.system_instruction = system_instruction
        self.example_field_values = example_field_values
        self.response_format = response_format
        self.output_fields = output_fields
        self.inference_engine = inference_engine

        # Validate the configuration
        if prompt_template is None or not prompt_template.strip():
            raise ValueError("Prompt template cannot be empty or None")
        self._validate_output_fields(output_fields)

    def judge(
        self,
        inputs: list[dict[str, str]],
    ) -> list[JudgeOutput]:
        """Evaluate a batch of inputs and return structured judgments.

        Args:
            inputs: List of dictionaries containing input data for evaluation.
                    Each dict must contain values for all prompt_template placeholders.

        Returns:
            List of structured judge outputs with parsed results

        Raises:
            ValueError: If inference returns unexpected number of conversations
        """
        # Fast fail if the dataset is invalid
        self.validate_dataset(inputs)

        # Build few-shot examples: convert field values to prompts and responses
        example_user_prompts = [
            self._build_judgment_prompt(example_fields)
            for example_fields in self.example_field_values
        ]
        example_assistant_responses = [
            self._build_assistant_response(example_fields)
            for example_fields in self.example_field_values
        ]

        # Build a judgment prompt for each dataset input
        judgment_prompts = [
            self._build_judgment_prompt(input_data) for input_data in inputs
        ]

        # Create a conversation for each judgment prompt
        judge_conversations = [
            self._build_judge_conversation(
                system_instruction=self.system_instruction,
                example_user_prompts=example_user_prompts,
                example_assistant_responses=example_assistant_responses,
                judgment_prompt=judgment_prompt,
            )
            for judgment_prompt in judgment_prompts
        ]

        # Run inference for all conversations in batch
        completed_conversations = self._infer(judge_conversations)

        # Extract and parse the judgment outputs
        judge_outputs = []
        for conversation in completed_conversations:
            self._validate_completed_conversation(conversation)

            raw_output = str(conversation.messages[-1].content)
            judge_output = self._transform_judge_output(raw_output)
            judge_outputs.append(judge_output)

        return judge_outputs

    def validate_dataset(
        self, inputs: list[dict[str, str]], raise_on_error: bool = True
    ) -> bool:
        """Validate that all inputs contain the required placeholder keys."""
        if self.prompt_template_placeholders is None:
            return True  # No validation needed if no placeholders are specified

        for index, input in enumerate(inputs):
            if missing_keys := self.prompt_template_placeholders - set(input.keys()):
                if raise_on_error:
                    raise ValueError(
                        f"Input {index} is missing keys: {sorted(missing_keys)}. "
                        f"Required: {sorted(self.prompt_template_placeholders)}, "
                        f"Found: {sorted(set(input.keys()))}."
                    )
                return False
        return True

    def _validate_output_fields(self, output_fields: list[JudgeOutputField]) -> None:
        """Ensure all output fields are properly defined."""
        if not output_fields:
            raise ValueError("Output fields cannot be empty")

        for field in self.output_fields:
            if field.field_key is None or not field.field_key.strip():
                raise ValueError(
                    f"Output field `field_key` cannot be None or empty: {field}"
                )
            if field.field_type == JudgeOutputType.ENUM and not field.field_scores:
                raise ValueError(
                    f"ENUM field type requires `field_scores` to be defined: {field}"
                )

    def _build_judgment_prompt(self, judge_input: dict[str, str]) -> str:
        """Generate a judge prompt by filling the template with input data.

        Args:
            judge_input: Dictionary mapping placeholder names to values

        Returns:
            Formatted prompt string ready for inference

        Raises:
            ValueError: If required placeholders are missing from judge_input
        """
        return resolve_placeholders(self.prompt_template, judge_input)

    def _build_assistant_response(self, field_values: dict[str, str]) -> str:
        """Generate the expected assistant response for few-shot examples.

        Args:
            field_values: Dictionary mapping field keys to their expected values.

        Returns:
            Formatted response string in the judge's expected output format.
        """
        # Create JudgeOutput instance with format configuration
        judge_output = JudgeOutput(
            raw_output="",
            response_format=self.response_format,
            output_fields=self.output_fields,
        )
        return judge_output.generate_raw_output(field_values=field_values)

    @classmethod
    def _build_judge_conversation(
        cls,
        system_instruction: Optional[str],
        example_user_prompts: list[str],
        example_assistant_responses: list[str],
        judgment_prompt: str,
    ) -> Conversation:
        """Create a conversation object.

        The conversation structure is:
            1. System instruction (if provided)
            2. Few-shot examples: alternating user prompts and assistant responses
            3. Final user prompt for the actual judgment

        Args:
            system_instruction: Optional system message to guide judge behavior
            example_user_prompts: List of example user prompts for few-shot learning
            example_assistant_responses: List of corresponding assistant responses
            judgment_prompt: The final user prompt for the actual judgment task

        Returns:
            Conversation object ready for inference

        Raises:
            ValueError: If number of example prompts doesn't match number of responses
        """
        # Validate that examples are properly paired
        if len(example_user_prompts) != len(example_assistant_responses):
            raise ValueError(
                f"Number of prompts ({len(example_user_prompts)}) must match "
                f"number of responses ({len(example_assistant_responses)})"
            )

        messages = []

        # Add system instruction, if provided
        if system_instruction:
            messages.append(Message(content=system_instruction, role=Role.SYSTEM))

        # Add few-shot examples as alternating user/assistant pairs
        for prompt, response in zip(example_user_prompts, example_assistant_responses):
            messages.extend(
                [
                    Message(content=prompt, role=Role.USER),
                    Message(content=response, role=Role.ASSISTANT),
                ]
            )

        # Add the actual judgment prompt
        messages.append(Message(content=judgment_prompt, role=Role.USER))

        return Conversation(messages=messages)

    def _validate_completed_conversation(self, conversation: Conversation) -> None:
        """Validate that completed conversation has the expected structure."""
        # Calculate expected message count
        expected_messages = 2  # Judgment user prompt + assistant response
        if self.system_instruction:
            expected_messages += 1
        expected_messages += 2 * len(self.example_field_values)

        if len(conversation.messages) != expected_messages:
            raise ValueError(
                f"Expected {expected_messages} messages, "
                f"got {len(conversation.messages)}"
            )

        # Validate that the last message is from the assistant
        if conversation.messages[-1].role != Role.ASSISTANT:
            raise ValueError(
                f"Expected last message from assistant, "
                f"got {conversation.messages[-1].role}"
            )

    def _infer(self, conversations: list[Conversation]) -> list[Conversation]:
        """Run inference on judge conversations and preserve metadata.

        Args:
            conversations: List of conversations to run inference on

        Returns:
            List of conversations with model responses added
        """
        # Preserve original metadata from input conversations
        original_metadata = [conv.metadata for conv in conversations]

        # Run batch inference
        response_conversations = self.inference_engine.infer(input=conversations)

        if len(response_conversations) != len(original_metadata):
            raise ValueError(
                f"Inference engine returned {len(response_conversations)} responses "
                f"but expected {len(original_metadata)}"
            )

        # Restore original metadata to response conversations
        for response_conv, metadata in zip(response_conversations, original_metadata):
            response_conv.metadata.update(metadata)

        return response_conversations

    def _transform_judge_output(self, raw_output: str) -> JudgeOutput:
        """Parse raw model output into structured judge output.

        Args:
            raw_output: The raw string output from the judge model

        Returns:
            Structured judge output with parsed fields and values
        """
        return JudgeOutput.from_raw_output(
            raw_output=raw_output,
            response_format=self.response_format,
            output_fields=self.output_fields,
        )



import os
from datetime import datetime
from pathlib import Path


def get_e2e_test_output_dir(test_name: str, tmp_path: Path) -> Path:
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    if os.environ.get("OUMI_E2E_TESTS_OUTPUT_DIR"):
        output_base = Path(os.environ["OUMI_E2E_TESTS_OUTPUT_DIR"])
    else:
        output_base = tmp_path / "e2e_tests"

    return output_base / (
        f"{timestamp}"  # Only timestamp if test_name is in output_base already.
        if f"/{test_name}/" in str(output_base)
        else f"{timestamp}_{test_name}"
    )


def is_file_not_empty(file_path: Path) -> bool:
    """Check if a file is not empty."""
    return file_path.stat().st_size > 0



import json
import os

import pytest
import torch
from transformers import AutoTokenizer

from oumi.core.datasets import BasePretrainingDataset


#
# Fixtures
#
@pytest.fixture
def tokenizer():
    return AutoTokenizer.from_pretrained("gpt2")


class TestDataset(BasePretrainingDataset):
    def __init__(self, *args, mock_data=None, **kwargs):
        self.mock_data = (
            mock_data if mock_data is not None else self._default_mock_data()
        )
        super().__init__(*args, **kwargs)

    def _load_data(self):
        return self.mock_data

    def _default_mock_data(self):
        return [
            {"text": "This is a test sentence."},
            {"text": "Another example sentence for testing."},
            {"text": "A third sentence to ensure we have enough data."},
        ]


@pytest.fixture
def test_dataset(tokenizer):
    return TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy_path",
        seq_length=10,
        dataset_text_field="text",
    )


@pytest.fixture
def create_sample_data(tmp_path):
    data_dir = tmp_path / "sample_data"
    data_dir.mkdir()

    for i in range(5):
        file_path = data_dir / f"data_{i}.json"
        with open(file_path, "w") as f:
            json.dump({"text": f"This is sample text number {i}. " * 10}, f)

    return str(data_dir)


class DiskDataset(BasePretrainingDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_dir = self.dataset_name

    def _load_data(self):
        for filename in os.listdir(self.data_dir):
            if filename.endswith(".json"):
                file_path = os.path.join(self.data_dir, filename)
                with open(file_path) as f:
                    yield json.load(f)


#
# Tests
#
def test_initialization(tokenizer):
    dataset = TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy_path",
        seq_length=10,
        dataset_text_field="text",
    )
    assert dataset.tokenizer == tokenizer
    assert dataset.seq_length == 10
    assert dataset._dataset_text_field == "text"


def test_tokenize(test_dataset, tokenizer):
    text = "Test sentence"
    tokens = test_dataset.tokenize(text)
    assert isinstance(tokens, list)
    assert len(tokens) > 0
    assert all(isinstance(token, int) for token in tokens)


def test_create_sample(test_dataset, tokenizer):
    tokens = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    sample = test_dataset._create_training_sample(tokens)
    assert isinstance(sample, dict)
    assert "input_ids" in sample
    assert "attention_mask" in sample
    assert "labels" in sample
    assert all(isinstance(tensor, torch.Tensor) for tensor in sample.values())
    assert all(tensor.shape == torch.Size([10]) for tensor in sample.values())


def test_iter(test_dataset, tokenizer):
    samples = list(test_dataset)
    assert len(samples) > 0
    for sample in samples:
        assert isinstance(sample, dict)
        assert "input_ids" in sample
        assert "attention_mask" in sample
        assert "labels" in sample
        assert all(isinstance(tensor, torch.Tensor) for tensor in sample.values())
        assert all(tensor.shape == torch.Size([10]) for tensor in sample.values())


def test_buffer_handling(tokenizer):
    dataset = TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy_path",
        seq_length=20,  # Longer sequence length to test buffer handling
        dataset_text_field="text",
    )
    samples = list(dataset)
    total_tokens = sum(
        len(dataset.tokenize(item["text"])) for item in dataset._load_data()
    )
    expected_samples = total_tokens // 20
    assert len(samples) == expected_samples


def test_disk_dataset(tokenizer, create_sample_data):
    dataset = DiskDataset(
        tokenizer=tokenizer,
        dataset_name=create_sample_data,
        seq_length=50,
        dataset_text_field="text",
    )

    samples = list(dataset)
    assert len(samples) > 0

    for sample in samples:
        assert isinstance(sample, dict)
        assert "input_ids" in sample
        assert "attention_mask" in sample
        assert "labels" in sample
        assert all(isinstance(tensor, torch.Tensor) for tensor in sample.values())
        assert all(tensor.shape == torch.Size([50]) for tensor in sample.values())

    # Check if we've processed all files
    num_files = len([f for f in os.listdir(create_sample_data) if f.endswith(".json")])
    total_tokens = sum(
        len(
            dataset.tokenize(
                json.load(open(os.path.join(create_sample_data, f)))["text"]
            )
        )
        for f in os.listdir(create_sample_data)
        if f.endswith(".json")
    )
    expected_samples = total_tokens // 50
    assert len(samples) == expected_samples
    assert len(samples) >= num_files  # At least one sample per file


def test_dataset_with_exact_sequence_length(tokenizer):
    mock_data = [
        {
            "text": "This sentence is exactly twenty tokens long sentence "
            "for testing purposes using gpt2's default tokenizer."
        }
    ]
    dataset = TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy",
        seq_length=20,
        mock_data=mock_data,
        skip_last=False,
    )
    samples = list(dataset)
    assert len(samples) == 1, "Dataset contains 20 tokens, we should have one sample"
    assert all(tensor.shape[0] == 20 for tensor in samples[0].values())

    dataset = TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy",
        seq_length=19,
        mock_data=mock_data,
        skip_last=False,
    )
    samples = list(dataset)
    assert len(samples) == 2
    assert samples[0]["input_ids"].shape[0] == 19, (
        "First sample should match seq_length"
    )
    assert samples[1]["input_ids"].shape[0] == 1, (
        "Second sample should have the remainder"
    )


def test_dataset_with_very_long_sequence(tokenizer):
    mock_data = [{"text": "long " * 1000}]  # 1000 tokens
    dataset = TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy",
        seq_length=10,
        mock_data=mock_data,
    )

    samples = list(dataset)
    assert len(samples) == 100
    assert all(
        tensor.shape[0] == 10 for sample in samples for tensor in sample.values()
    )


def test_dataset_with_non_text_field(tokenizer):
    mock_data = [{"non_text": "This should not be processed."}]
    dataset = TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy",
        seq_length=10,
        dataset_text_field="text",
        mock_data=mock_data,
    )
    with pytest.raises(KeyError):
        list(dataset)


def test_dataset_with_no_docs(tokenizer):
    mock_data = []
    dataset = TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy",
        seq_length=10,
        dataset_text_field="text",
        mock_data=mock_data,
        skip_last=False,
    )

    samples = list(dataset)
    assert len(samples) == 0


def test_dataset_with_empty_docs(tokenizer):
    mock_data = [{"text": ""}, {"text": ""}, {"text": ""}, {"text": ""}, {"text": ""}]
    dataset = TestDataset(
        tokenizer=tokenizer,
        dataset_name="dummy",
        seq_length=10,
        dataset_text_field="text",
        mock_data=mock_data,
        skip_last=False,
    )

    samples = list(dataset)
    assert len(samples) == 0



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from pathlib import Path
from unittest.mock import MagicMock, Mock, patch

import jsonlines
import pandas as pd
import pytest

from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.datasets.vision_language.vision_dpo_jsonlines import VisionDpoJsonlinesDataset

_IMAGE_TOKEN = "<image_token>"
_IMAGE_TOKEN_ID = 32001


@pytest.fixture
def mock_processor():
    processor = Mock()
    processor.processor_name = "llava-hf/llava-1.5-7b-hf"
    processor.tokenizer = Mock()
    processor.image_processor = Mock()
    processor.image_processor.size = {"longest_edge": 224}
    processor.chat_template = None
    processor.image_token = _IMAGE_TOKEN
    processor.image_token_id = _IMAGE_TOKEN_ID
    processor.label_ignore_index = -100
    return processor


@pytest.fixture
def mock_tokenizer() -> MagicMock:
    def _convert_tokens_to_ids(token: str) -> int:
        if token == _IMAGE_TOKEN:
            return _IMAGE_TOKEN_ID
        return 101

    mock = MagicMock(spec=BaseTokenizer)
    mock.pad_token_id = 0
    mock.convert_tokens_to_ids = MagicMock(side_effect=_convert_tokens_to_ids)
    return mock


@pytest.fixture
def sample_vision_dpo_data():
    return [
        {
            "prompt": "What do you see in this image?",
            "images": ["https://picsum.photos/200/300"],
            "chosen": [
                {"role": "assistant", "content": "I see a beautiful landscape."}
            ],
            "rejected": [{"role": "assistant", "content": "I see nothing."}],
        },
        {
            "prompt": "Describe the main subject of this image.",
            "images": ["path/to/local/image.jpg"],
            "chosen": [
                {
                    "role": "assistant",
                    "content": "The main subject is a golden retriever dog.",
                }
            ],
            "rejected": [
                {"role": "assistant", "content": "The main subject is a cat."}
            ],
        },
        {
            "prompt": "What activity is happening in this image?",
            "images": [],
            "chosen": [
                {
                    "role": "assistant",
                    "content": "The image shows children playing soccer.",
                }
            ],
            "rejected": [
                {"role": "assistant", "content": "The image shows people swimming."}
            ],
        },
    ]


def test_init_with_data(sample_vision_dpo_data, mock_tokenizer, mock_processor):
    dataset = VisionDpoJsonlinesDataset(
        data=sample_vision_dpo_data, tokenizer=mock_tokenizer, processor=mock_processor
    )

    assert len(dataset._data) == 3
    assert isinstance(dataset._data, pd.DataFrame)
    assert list(dataset._data.columns) == ["prompt", "images", "chosen", "rejected"]

    # Check first row
    first_row = dataset._data.iloc[0]
    assert first_row["prompt"] == "What do you see in this image?"
    assert first_row["images"] == ["https://picsum.photos/200/300"]
    assert first_row["chosen"] == [
        {"role": "assistant", "content": "I see a beautiful landscape."}
    ]
    assert first_row["rejected"] == [{"role": "assistant", "content": "I see nothing."}]


def test_init_with_dataset_path(sample_vision_dpo_data, mock_tokenizer, mock_processor):
    with tempfile.TemporaryDirectory() as folder:
        jsonl_file = Path(folder) / "vision_dpo_test.jsonl"
        with jsonlines.open(jsonl_file, mode="w") as writer:
            writer.write_all(sample_vision_dpo_data)

        dataset = VisionDpoJsonlinesDataset(
            dataset_path=str(jsonl_file),
            tokenizer=mock_tokenizer,
            processor=mock_processor,
        )

        assert len(dataset._data) == 3
        assert isinstance(dataset._data, pd.DataFrame)

        # Check that data was loaded correctly
        first_row = dataset._data.iloc[0]
        assert first_row["prompt"] == "What do you see in this image?"
        assert first_row["images"] == ["https://picsum.photos/200/300"]


def test_init_validation_errors(sample_vision_dpo_data, mock_tokenizer, mock_processor):
    # Test missing both dataset_path and data
    with pytest.raises(
        ValueError, match="Either dataset_path or data must be provided"
    ):
        VisionDpoJsonlinesDataset(tokenizer=mock_tokenizer, processor=mock_processor)

    # Test providing both dataset_path and data
    with tempfile.TemporaryDirectory() as folder:
        jsonl_file = Path(folder) / "test.jsonl"
        jsonl_file.write_text('{"test": "data"}')

        with pytest.raises(
            ValueError, match="Only one of dataset_path or data must be provided"
        ):
            VisionDpoJsonlinesDataset(
                dataset_path=str(jsonl_file),
                data=sample_vision_dpo_data,
                tokenizer=mock_tokenizer,
                processor=mock_processor,
            )


def test_transform_preference(sample_vision_dpo_data, mock_tokenizer, mock_processor):
    dataset = VisionDpoJsonlinesDataset(
        data=sample_vision_dpo_data, tokenizer=mock_tokenizer, processor=mock_processor
    )

    sample = sample_vision_dpo_data[0]

    with (
        patch.object(dataset, "_load_image") as mock_load_image,
        patch.object(dataset, "_resize_image") as mock_resize_image,
    ):
        mock_image = Mock()
        mock_load_image.return_value = mock_image
        mock_resize_image.return_value = mock_image

        result = dataset.transform_preference(sample)

        assert "prompt" in result
        assert "chosen" in result
        assert "rejected" in result
        assert "images" in result



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from typing import Optional

import PIL.Image
import transformers


class BaseImageProcessor(abc.ABC):
    """Base class for oumi image processors."""

    @abc.abstractmethod
    def __call__(
        self,
        images: list[PIL.Image.Image],
        *,
        return_tensors: Optional[str] = "pt",
    ) -> transformers.BatchFeature:
        """Extracts image features.

        Args:
            images: A list of input images.
            return_tensors: The format of returned tensors.

        Returns:
            transformers.BatchFeature: The model-specific input features.
        """
        raise NotImplementedError



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

from oumi.utils.logging import logger


class AliasType(str, Enum):
    """The type of configs we support with aliases."""

    TRAIN = "train"
    EVAL = "eval"
    INFER = "infer"
    JOB = "job"
    QUANTIZE = "quantize"
    JUDGE = "judge"
    TUNE = "tune"


_ALIASES: dict[str, dict[AliasType, str]] = {
    # Llama 4 family.
    "llama4-scout": {
        AliasType.TRAIN: "oumi://configs/recipes/llama4/sft/scout_base_full/train.yaml",
        AliasType.JOB: "oumi://configs/recipes/llama4/sft/scout_base_full/gcp_job.yaml",
    },
    "llama4-scout-instruct-lora": {
        AliasType.TRAIN: "oumi://configs/recipes/llama4/sft/scout_instruct_lora/train.yaml",
    },
    "llama4-scout-instruct-qlora": {
        AliasType.TRAIN: "oumi://configs/recipes/llama4/sft/scout_instruct_qlora/train.yaml",
    },
    "llama4-scout-instruct": {
        AliasType.TRAIN: "oumi://configs/recipes/llama4/sft/scout_instruct_full/train.yaml",
        AliasType.INFER: "oumi://configs/recipes/llama4/inference/scout_instruct_infer.yaml",
        AliasType.JOB: "oumi://configs/recipes/llama4/sft/scout_instruct_full/gcp_job.yaml",
        AliasType.EVAL: "oumi://configs/recipes/llama4/evaluation/scout_instruct_eval.yaml",
    },
    "llama4-maverick": {
        AliasType.INFER: "oumi://configs/recipes/llama4/inference/maverick_instruct_together_infer.yaml",
    },
    # Qwen3 family.
    "qwen3-30b-a3b": {
        AliasType.INFER: "oumi://configs/recipes/qwen3/inference/30b_a3b_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/qwen3/evaluation/30b_a3b_eval.yaml",
    },
    "qwen3-30b-a3b-lora": {
        AliasType.TRAIN: "oumi://configs/recipes/qwen3/sft/30b_a3b_lora/train.yaml",
        AliasType.JOB: "oumi://configs/recipes/qwen3/sft/30b_a3b_lora/gcp_job.yaml",
    },
    "qwen3-32b": {
        AliasType.INFER: "oumi://configs/recipes/qwen3/inference/32b_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/qwen3/evaluation/32b_eval.yaml",
    },
    "qwen3-32b-lora": {
        AliasType.TRAIN: "oumi://configs/recipes/qwen3/sft/32b_lora/train.yaml",
        AliasType.JOB: "oumi://configs/recipes/qwen3/sft/32b_lora/gcp_job.yaml",
    },
    # Phi family.
    "phi4-reasoning-plus": {
        AliasType.TRAIN: "oumi://configs/recipes/phi4/sft/reasoning_plus/full_train.yaml",
        AliasType.JOB: "oumi://configs/recipes/phi4/sft/reasoning_plus/full_gcp_job.yaml",
        AliasType.INFER: "oumi://configs/recipes/phi4/inference/reasoning_plus_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/phi4/evaluation/reasoning_plus_eval.yaml",
    },
    "phi4-reasoning-plus-lora": {
        AliasType.TRAIN: "oumi://configs/recipes/phi4/sft/reasoning_plus/lora_train.yaml",
        AliasType.JOB: "oumi://configs/recipes/phi4/sft/reasoning_plus/lora_gcp_job.yaml",
    },
    "phi4-reasoning-plus-qlora": {
        AliasType.TRAIN: "oumi://configs/recipes/phi4/sft/reasoning_plus/qlora_train.yaml",
        AliasType.JOB: "oumi://configs/recipes/phi4/sft/reasoning_plus/qlora_gcp_job.yaml",
    },
    # Falcon H1 family
    "falcon_h1_0_5b": {
        AliasType.TRAIN: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_0_5b/full_train.yaml",
        AliasType.INFER: "oumi://configs/recipes/falcon_h1/inference/0_5b_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/falcon_h1/evaluation/falcon_h1_0_5b/eval.yaml",
        AliasType.JOB: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_0_5b/full_gcp_job.yaml",
    },
    "falcon_h1_1_5b": {
        AliasType.TRAIN: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_1_5b/full_train.yaml",
        AliasType.INFER: "oumi://configs/recipes/falcon_h1/inference/1_5b_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/falcon_h1/evaluation/falcon_h1_1_5b/eval.yaml",
        AliasType.JOB: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_1_5b/full_gcp_job.yaml",
    },
    "falcon_h1_1_5b_deep": {
        AliasType.TRAIN: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_1_5b_deep/full_train.yaml",
        AliasType.INFER: "oumi://configs/recipes/falcon_h1/inference/1_5b_deep_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/falcon_h1/evaluation/falcon_h1_1_5b_deep/eval.yaml",
        AliasType.JOB: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_1_5b_deep/full_gcp_job.yaml",
    },
    "falcon_h1_3b": {
        AliasType.TRAIN: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_3b/full_train.yaml",
        AliasType.INFER: "oumi://configs/recipes/falcon_h1/inference/3b_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/falcon_h1/evaluation/falcon_h1_3b/eval.yaml",
        AliasType.JOB: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_3b/full_gcp_job.yaml",
    },
    "falcon_h1_7b": {
        AliasType.TRAIN: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_7b/full_train.yaml",
        AliasType.INFER: "oumi://configs/recipes/falcon_h1/inference/7b_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/falcon_h1/evaluation/falcon_h1_7b/eval.yaml",
        AliasType.JOB: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_7b/full_gcp_job.yaml",
    },
    "falcon_h1_34b": {
        AliasType.TRAIN: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_34b/full_train.yaml",
        AliasType.INFER: "oumi://configs/recipes/falcon_h1/inference/34b_infer.yaml",
        AliasType.EVAL: "oumi://configs/recipes/falcon_h1/evaluation/falcon_h1_34b/eval.yaml",
        AliasType.JOB: "oumi://configs/recipes/falcon_h1/sft/falcon_h1_34b/full_gcp_job.yaml",
    },
    # Hosted models.
    "claude-3-5-sonnet": {
        AliasType.INFER: "oumi://configs/apis/anthropic/infer_claude_3_5_sonnet.yaml",
        AliasType.EVAL: "oumi://configs/apis/anthropic/eval_claude_3_5_sonnet.yaml",
    },
    "claude-3-7-sonnet": {
        AliasType.INFER: "oumi://configs/apis/anthropic/infer_claude_3_7_sonnet.yaml",
        AliasType.EVAL: "oumi://configs/apis/anthropic/eval_claude_3_7_sonnet.yaml",
    },
    "gemini-1-5-pro": {
        AliasType.INFER: "oumi://configs/apis/gemini/infer_gemini_1_5_pro.yaml",
        AliasType.EVAL: "oumi://configs/apis/gemini/eval_gemini_1_5_pro.yaml",
    },
    "gpt-4o": {
        AliasType.INFER: "oumi://configs/apis/openai/infer_gpt_4o.yaml",
        AliasType.EVAL: "oumi://configs/apis/openai/eval_gpt_4o.yaml",
    },
    "gpt-o1-preview": {
        AliasType.INFER: "oumi://configs/apis/openai/infer_gpt_o1_preview.yaml",
        AliasType.EVAL: "oumi://configs/apis/openai/eval_gpt_o1_preview.yaml",
    },
    "llama-3-3-70b": {
        AliasType.INFER: "oumi://configs/apis/vertex/infer_llama_3_3_70b.yaml",
        AliasType.EVAL: "oumi://configs/apis/vertex/eval_llama_3_3_70b.yaml",
    },
    "llama-3-1-405b": {
        AliasType.INFER: "oumi://configs/apis/vertex/infer_llama_3_1_405b.yaml",
        AliasType.EVAL: "oumi://configs/apis/vertex/eval_llama_3_1_405b.yaml",
    },
    "molmo-7b-o": {
        AliasType.TRAIN: "oumi://configs/recipes/vision/molmo/sft/molmo_o_full/train.yaml",
    },
    "molmo-7b-d": {
        AliasType.TRAIN: "oumi://configs/recipes/vision/molmo/sft/molmo_d_full/train.yaml",
    },
    "molmo-7b-o-grpo": {
        AliasType.TRAIN: "oumi://configs/recipes/vision/molmo/grpo/train.yaml",
    },
    # Judge aliases (for generic judges only)
    "format-compliance": {
        AliasType.JUDGE: "oumi://configs/projects/judges/generic/format_compliance.yaml",
    },
    "instruction-following": {
        AliasType.JUDGE: "oumi://configs/projects/judges/generic/instruction_following.yaml",
    },
    "topic-adherence": {
        AliasType.JUDGE: "oumi://configs/projects/judges/generic/topic_adherence.yaml",
    },
    "truthfulness": {
        AliasType.JUDGE: "oumi://configs/projects/judges/generic/truthfulness.yaml",
    },
    "safety": {
        AliasType.JUDGE: "oumi://configs/projects/judges/generic/safety.yaml",
    },
}


def try_get_config_name_for_alias(
    alias: str,
    alias_type: AliasType,
) -> str:
    """Gets the config path for a given alias.

    This function resolves the config path for a given alias and alias type.
    If the alias is not found, the original alias is returned.

    Args:
        alias (str): The alias to resolve.
        alias_type (AliasType): The type of config to resolve.

    Returns:
        str: The resolved config path (or the original alias if not found).
    """
    if alias in _ALIASES and alias_type in _ALIASES[alias]:
        config_path = _ALIASES[alias][alias_type]
        logger.info(f"Resolved alias '{alias}' to '{config_path}'")
        return config_path
    return alias



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional

import numpy as np
from sklearn.metrics import balanced_accuracy_score
from sklearn.metrics import f1_score as sklearn_f1_score
from sklearn.utils import resample


class Metric:
    """Base class for all metrics."""

    name: Optional[str] = ""
    """The name of the metric."""

    value: Optional[float] = None
    """The value of the metric."""

    ci: Optional[float] = None
    """The confidence interval of the metric."""

    def __init__(
        self,
        name: Optional[str] = "",
        value: Optional[float] = None,
        ci: Optional[float] = None,
    ) -> None:
        """Initialize a Metric instance."""
        self.name = name
        self.value = value
        self.ci = ci

    def __str__(self):
        """Return a string representation of the Metric instance."""
        return f"Metric(name={self.name}, value={self.value}, ci={self.ci})"


def bootstrap(
    y_true: list[int],
    y_pred: list[int],
    metric_fn: Callable[..., float],
    alpha: float = 0.95,
    n_iter: int = 1000,
    sample_prop=1.0,
) -> tuple[float, float]:
    """Perform bootstrap resampling to calculate confidence intervals for a metric.

    Args:
        y_true (list): True labels.
        y_pred (list): Predicted labels.
        metric_fn (callable): A function that computes a performance metric. The
            required signature for this function is:
            `metric_fn(y_true: list[int], y_pred: list[int]) -> float`.
        alpha (float): Confidence level (default is 0.95 for 95% confidence interval).
        n_iter (int): Number of bootstrap iterations (default is 1000).
        sample_prop (float): Proportion of the data to sample in each iteration
            (default 1.0, meaning same size as original).

    Returns:
        tuple: Midpoint and half-width of the confidence interval.
    """
    # Ensure that our inputs are correct.
    if not y_true or not y_pred:
        raise ValueError("`y_true` and `y_pred` must not be empty.")
    if len(y_true) != len(y_pred):
        raise ValueError("`y_true` and `y_pred` must have the same length.")
    if not (0 < alpha < 1):
        raise ValueError("`alpha` must be between 0 and 1.")
    if n_iter <= 0:
        raise ValueError("`n_iter` must be a positive integer.")
    if not (0 < sample_prop <= 1):
        raise ValueError("`sample_prop` must be between 0 and 1.")

    # Combine the true values and predictions into paired data points.
    data = list(zip(y_true, y_pred))

    # Determine the sample size for each bootstrap iteration.
    n_size = int(len(data) * sample_prop)

    # Run bootstrap.
    scores = []  # Metric scores calculated during the bootstrap iterations.
    for _ in range(n_iter):
        samples = resample(data, n_samples=n_size)
        if not samples:  # samples should never be None.
            continue
        labels = [sample[0] for sample in samples]
        predictions = [sample[1] for sample in samples]
        score = metric_fn(y_true=labels, y_pred=predictions)
        scores.append(score)

    # Calculate the confidence interval.
    lower_percentile = ((1.0 - alpha) / 2.0) * 100
    upper_percentile = (alpha + ((1.0 - alpha) / 2.0)) * 100

    lower_bound = max(0.0, float(np.percentile(scores, lower_percentile)))
    upper_bound = min(1.0, float(np.percentile(scores, upper_percentile)))
    midpoint = (upper_bound + lower_bound) / 2.0

    return midpoint, upper_bound - midpoint


def f1_score(
    y_true: list[int],
    y_pred: list[int],
    average: str = "binary",
    pos_label: int = 1,
    populate_ci: bool = True,
    alpha: float = 0.95,
    n_iter: int = 1000,
    sample_prop=1.0,
) -> Metric:
    """Calculate the F1 score and its confidence interval."""
    # Ensure that our inputs are correct.
    if not y_true or not y_pred:
        raise ValueError("`y_true` and `y_pred` must not be empty.")
    if len(y_true) != len(y_pred):
        raise ValueError("`y_true` and `y_pred` must have the same length.")
    if average not in ["micro", "macro", "samples", "weighted", "binary"]:
        raise ValueError("Invalid value for `average`.")
    if not (0 < alpha < 1):
        raise ValueError("`alpha` must be between 0 and 1.")
    if n_iter <= 0:
        raise ValueError("`n_iter` must be a positive integer.")
    if not (0 < sample_prop <= 1):
        raise ValueError("`sample_prop` must be between 0 and 1.")

    def f1_fn(y_true: list[int], y_pred: list[int]) -> float:
        return float(
            sklearn_f1_score(
                y_true=y_true,
                y_pred=y_pred,
                pos_label=pos_label,
                average=average,
            )
        )

    # Calculate F1 score
    if populate_ci:
        f1, ci = bootstrap(
            y_true=y_true,
            y_pred=y_pred,
            metric_fn=f1_fn,
            alpha=alpha,
            n_iter=n_iter,
            sample_prop=sample_prop,
        )
    else:
        f1 = f1_fn(
            y_true=y_true,
            y_pred=y_pred,
        )
        ci = None

    return Metric(name="F1 Score", value=f1, ci=ci)


def bacc_score(
    y_true: list[int],
    y_pred: list[int],
    populate_ci: bool = True,
    alpha: float = 0.95,
    n_iter: int = 1000,
    sample_prop=1.0,
) -> Metric:
    """Calculate the balanced accuracy score."""
    # Ensure that our inputs are correct.
    if not y_true or not y_pred:
        raise ValueError("`y_true` and `y_pred` must not be empty.")
    if len(y_true) != len(y_pred):
        raise ValueError("`y_true` and `y_pred` must have the same length.")
    if not (0 < alpha < 1):
        raise ValueError("`alpha` must be between 0 and 1.")
    if n_iter <= 0:
        raise ValueError("`n_iter` must be a positive integer.")
    if not (0 < sample_prop <= 1):
        raise ValueError("`sample_prop` must be between 0 and 1.")

    if populate_ci:
        # Calculate balanced accuracy score with bootstrap
        bacc, ci = bootstrap(
            y_true=y_true,
            y_pred=y_pred,
            metric_fn=balanced_accuracy_score,
            alpha=alpha,
            n_iter=n_iter,
            sample_prop=sample_prop,
        )
    else:
        bacc = balanced_accuracy_score(
            y_true=y_true,
            y_pred=y_pred,
        )
        ci = None

    return Metric(name="Balanced Accuracy", value=bacc, ci=ci)
